# -*- coding: utf-8 -*-
import base64
import datetime
import hashlib
import json
import warnings

import mock
import pytest

from ddtrace.internal.compat import PY2
from ddtrace.internal.remoteconfig import RemoteConfig
from ddtrace.internal.remoteconfig.client import RemoteConfigClient
from ddtrace.internal.remoteconfig.constants import ASM_FEATURES_PRODUCT
from ddtrace.internal.remoteconfig.constants import REMOTE_CONFIG_AGENT_ENDPOINT
from ddtrace.internal.remoteconfig.worker import RemoteConfigWorker
from ddtrace.internal.remoteconfig.worker import get_poll_interval_seconds
from tests.internal.test_utils_version import _assert_and_get_version_agent_format
from tests.utils import override_env


def to_bytes(string):
    if PY2:
        return bytes(string)
    else:
        return bytes(string, encoding="utf-8")


def to_str(bytes_string):
    if PY2:
        return str(bytes_string)
    else:
        return str(bytes_string, encoding="utf-8")


def get_mock_encoded_msg(msg):
    expires_date = datetime.datetime.strftime(
        datetime.datetime.now() + datetime.timedelta(days=1), "%Y-%m-%dT%H:%M:%SZ"
    )
    path = "datadog/2/%s/asm_features_activation/config" % ASM_FEATURES_PRODUCT
    data = {
        "signatures": [{"keyid": "", "sig": ""}],
        "signed": {
            "_type": "targets",
            "custom": {"opaque_backend_state": ""},
            "expires": expires_date,
            "spec_version": "1.0.0",
            "targets": {
                path: {
                    "custom": {"c": [""], "v": 0},
                    "hashes": {"sha256": hashlib.sha256(msg).hexdigest()},
                    "length": 24,
                }
            },
            "version": 0,
        },
    }
    return {
        "roots": [
            to_str(
                base64.b64encode(
                    to_bytes(
                        json.dumps(
                            {
                                "signatures": [],
                                "signed": {
                                    "_type": "root",
                                    "consistent_snapshot": True,
                                    "expires": "1986-12-11T00:00:00Z",
                                    "keys": {},
                                    "roles": {},
                                    "spec_version": "1.0",
                                    "version": 2,
                                },
                            }
                        ),
                    )
                )
            )
        ],
        "targets": to_str(base64.b64encode(to_bytes(json.dumps(data)))),
        "target_files": [
            {
                "path": path,
                "raw": to_str(base64.b64encode(msg)),
            }
        ],
        "client_configs": [path],
    }


def test_remote_config_register_auto_enable():
    # ASM_FEATURES product is enabled by default, but LIVE_DEBUGGER isn't
    with override_env(dict(DD_REMOTE_CONFIGURATION_ENABLED="true")):
        assert RemoteConfig._worker is None

        RemoteConfig.register("LIVE_DEBUGGER", lambda m, c: None)

        assert RemoteConfig._worker._client._products["LIVE_DEBUGGER"] is not None

        RemoteConfig.disable()


def test_remote_config_register_validate_rc_disabled():
    assert RemoteConfig._worker is None

    with override_env(dict(DD_REMOTE_CONFIGURATION_ENABLED="false")):
        RemoteConfig.register("LIVE_DEBUGGER", lambda m, c: None)

        assert RemoteConfig._worker is None


def test_remote_config_enable_validate_rc_disabled():
    assert RemoteConfig._worker is None

    with override_env(dict(DD_REMOTE_CONFIGURATION_ENABLED="false")):
        RemoteConfig.enable()

        assert RemoteConfig._worker is None


@pytest.mark.subprocess
def test_remote_config_forksafe():
    import os

    from ddtrace.internal.remoteconfig import RemoteConfig
    from tests.utils import override_env

    with override_env(dict(DD_REMOTE_CONFIGURATION_ENABLED="true")):
        RemoteConfig.enable()

        parent_worker = RemoteConfig._worker
        assert parent_worker is not None

        if os.fork() == 0:
            assert RemoteConfig._worker is not None
            assert RemoteConfig._worker is not parent_worker
            exit(0)


@mock.patch.object(RemoteConfigClient, "_send_request")
def test_remote_configuration_1_click(mock_send_request):
    class Callback:
        features = {}

        def _reload_features(self, metadata, features):
            self.features = features

    callback = Callback()
    with override_env(dict(DD_REMOTE_CONFIGURATION_ENABLED="true")):
        with RemoteConfig() as rc:
            mock_send_request.return_value = get_mock_encoded_msg(b'{"asm":{"enabled":true}}')
            rc.register(ASM_FEATURES_PRODUCT, callback._reload_features)
            rc._worker._online()
            mock_send_request.assert_called()
            assert callback.features == {"asm": {"enabled": True}}


def test_remote_configuration_check_deprecated_var():
    with override_env(dict(DD_REMOTE_CONFIG_POLL_INTERVAL_SECONDS="0.1")):
        with warnings.catch_warnings(record=True) as capture:
            get_poll_interval_seconds()
            assert len(capture) == 0


def test_remote_configuration_check_deprecated_var_message():
    with override_env(dict(DD_REMOTECONFIG_POLL_SECONDS="0.1")):
        with warnings.catch_warnings(record=True) as capture:
            get_poll_interval_seconds()
            assert len(capture) == 1
            assert str(capture[0].message).startswith("Using environment")


def test_remote_configuration_check_deprecated_override():
    with override_env(dict(DD_REMOTE_CONFIG_POLL_INTERVAL_SECONDS="0.1", DD_REMOTECONFIG_POLL_SECONDS="0.5")):
        with warnings.catch_warnings(record=True) as capture:
            assert get_poll_interval_seconds() == 0.1
            assert len(capture) == 1
            assert str(capture[0].message).startswith("Using environment")


@mock.patch.object(RemoteConfigClient, "_send_request")
def test_remote_configuration_ip_blocking(mock_send_request):
    class Callback:
        features = {}

        def _reload_features(self, metadata, features):
            self.features = features

    callback = Callback()

    with override_env(dict(DD_REMOTE_CONFIGURATION_ENABLED="true", DD_REMOTE_CONFIG_POLL_INTERVAL_SECONDS="0.1")):
        mock_send_request.return_value = get_mock_encoded_msg(
            b'{"rules_data": [{"data": [{"expiration": 1662804872, "value": "127.0.0.0"}, '
            b'{"expiration": 1662804872, "value": "52.80.198.1"}], "id": "blocking_ips", '
            b'"type": "ip_with_expiration"}]}'
        )
        with RemoteConfig() as rc:
            rc.register(ASM_FEATURES_PRODUCT, callback._reload_features)
            rc._worker._online()
            assert callback.features == {
                "rules_data": [
                    {
                        "data": [
                            {"expiration": 1662804872, "value": "127.0.0.0"},
                            {"expiration": 1662804872, "value": "52.80.198.1"},
                        ],
                        "id": "blocking_ips",
                        "type": "ip_with_expiration",
                    }
                ]
            }


def test_remoteconfig_semver():
    _assert_and_get_version_agent_format(RemoteConfigClient()._client_tracer["tracer_version"])


@pytest.mark.parametrize(
    "result,expected",
    [
        (None, False),
        ({}, False),
        ({"endpoints": []}, False),
        ({"endpoints": ["/info"]}, False),
        ({"endpoints": ["/info", "/errors"]}, False),
        ({"endpoints": ["/info", "/errors", REMOTE_CONFIG_AGENT_ENDPOINT]}, True),
        ({"endpoints": ["/info", "/errors", "/" + REMOTE_CONFIG_AGENT_ENDPOINT]}, True),
    ],
)
@mock.patch("ddtrace.internal.agent.info")
def test_remote_configuration_check_remote_config_enable_in_agent_errors(mock_info, result, expected):
    mock_info.return_value = result

    worker = RemoteConfigWorker()

    # Check that the initial state is agent_check
    assert worker._state == worker._agent_check

    worker.periodic()

    # Check that the state is online if the agent supports remote config
    assert worker._state == worker._online if expected else worker._agent_check



from time import sleep
from uuid import uuid4

import pytest

from ddtrace.debugging._config import config
from ddtrace.debugging._probe.model import DEFAULT_PROBE_RATE
from ddtrace.debugging._probe.model import DEFAULT_SNAPSHOT_PROBE_RATE
from ddtrace.debugging._probe.model import LogProbeMixin
from ddtrace.debugging._probe.model import Probe
from ddtrace.debugging._probe.model import ProbeType
from ddtrace.debugging._probe.remoteconfig import ProbePollerEvent
from ddtrace.debugging._probe.remoteconfig import ProbeRCAdapter
from ddtrace.debugging._probe.remoteconfig import _filter_by_env_and_version
from ddtrace.debugging._probe.remoteconfig import probe_factory
from ddtrace.internal.remoteconfig.client import ConfigMetadata
from tests.debugging.utils import create_snapshot_line_probe
from tests.utils import override_global_config


class MockConfig(object):
    def __init__(self, *args, **kwargs):
        self.probes = {}

    @_filter_by_env_and_version
    def get_probes(self, _id, _conf):
        return list(self.probes.values())

    def add_probes(self, probes):
        for probe in probes:
            self.probes[probe.probe_id] = probe

    def remove_probes(self, *probe_ids):
        for probe_id in probe_ids:
            try:
                del self.probes[probe_id]
            except KeyError:
                pass


def config_metadata(config_id=uuid4()):
    return ConfigMetadata(config_id, product_name="LIVE_DEBUGGING", sha256_hash="hash", length=123, tuf_version=1)


@pytest.fixture
def mock_config():
    import ddtrace.debugging._probe.remoteconfig as rc

    original_get_probes = rc.get_probes
    mock_config = MockConfig()
    try:
        rc.get_probes = mock_config.get_probes

        yield mock_config
    finally:
        rc.get_probes = original_get_probes


@pytest.mark.parametrize(
    "env,version,expected",
    [
        (None, None, set(["probe4"])),
        (None, "dev", set(["probe3", "probe4"])),
        ("prod", None, set(["probe2", "probe4"])),
        ("prod", "dev", set(["probe1", "probe2", "probe3", "probe4"])),
    ],
)
def test_poller_env_version(env, version, expected, mock_config):
    probes = []

    def cb(e, ps):
        probes.extend(ps)

    with override_global_config(dict(env=env, version=version)):
        mock_config.add_probes(
            [
                create_snapshot_line_probe(
                    probe_id="probe1",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                    tags={"env": "prod", "version": "dev"},
                ),
                create_snapshot_line_probe(
                    probe_id="probe2",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                    tags={"env": "prod"},
                ),
                create_snapshot_line_probe(
                    probe_id="probe3",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                    tags={"version": "dev"},
                ),
                create_snapshot_line_probe(
                    probe_id="probe4",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                ),
            ]
        )

        ProbeRCAdapter(cb)(config_metadata(), {})

        assert set(_.probe_id for _ in probes) == expected


def test_poller_events(mock_config):
    events = set()

    def cb(e, ps):
        events.add((e, frozenset([p.probe_id if isinstance(p, Probe) else p for p in ps])))

    mock_config.add_probes(
        [
            create_snapshot_line_probe(
                probe_id="probe1",
                source_file="tests/debugger/submod/stuff.py",
                line=36,
                condition=None,
            ),
            create_snapshot_line_probe(
                probe_id="probe2",
                source_file="tests/debugger/submod/stuff.py",
                line=36,
                condition=None,
            ),
            create_snapshot_line_probe(
                probe_id="probe3",
                source_file="tests/debugger/submod/stuff.py",
                line=36,
                condition=None,
            ),
            create_snapshot_line_probe(
                probe_id="probe4",
                source_file="tests/debugger/submod/stuff.py",
                line=36,
                condition=None,
            ),
        ]
    )

    metadata = config_metadata()
    old_interval = config.diagnostics_interval
    config.diagnostics_interval = 0.5
    try:
        adapter = ProbeRCAdapter(cb)

        adapter(metadata, {})
        mock_config.remove_probes("probe1", "probe2")
        mock_config.add_probes(
            [
                # Modified
                create_snapshot_line_probe(
                    probe_id="probe2",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                ),
                # New
                create_snapshot_line_probe(
                    probe_id="probe5",
                    source_file="tests/debugger/submod/stuff.py",
                    line=36,
                    condition=None,
                ),
            ]
        )
        adapter(metadata, {})

        # Wait to allow the next call to the adapter to generate a status event
        sleep(0.5)
        adapter(metadata, {})

        assert events == {
            (ProbePollerEvent.NEW_PROBES, frozenset(["probe4", "probe1", "probe2", "probe3"])),
            (ProbePollerEvent.DELETED_PROBES, frozenset(["probe1"])),
            (ProbePollerEvent.NEW_PROBES, frozenset(["probe5"])),
            (ProbePollerEvent.STATUS_UPDATE, frozenset(["probe4", "probe2", "probe3", "probe5"])),
        }
    finally:
        config.diagnostics_interval = old_interval


def test_multiple_configs():
    events = set()

    def cb(e, ps):
        events.add((e, frozenset({p.probe_id if isinstance(p, Probe) else p for p in ps})))

    def validate_events(expected):
        assert events == expected
        events.clear()

    old_interval = config.diagnostics_interval
    config.diagnostics_interval = 0.5
    try:
        adapter = ProbeRCAdapter(cb)

        adapter(
            config_metadata("metricProbe_probe2"),
            {
                "id": "probe2",
                "type": ProbeType.METRIC_PROBE,
                "tags": ["foo:bar"],
                "where": {"sourceFile": "tests/submod/stuff.p", "lines": ["36"]},
                "metricName": "test.counter",
                "kind": "COUNTER",
            },
        )

        validate_events(
            {
                (ProbePollerEvent.NEW_PROBES, frozenset({"probe2"})),
            }
        )

        adapter(
            config_metadata("logProbe_probe3"),
            {
                "id": "probe3",
                "type": ProbeType.LOG_PROBE,
                "tags": ["foo:bar"],
                "where": {"sourceFile": "tests/submod/stuff.p", "lines": ["36"]},
                "template": "hello {#foo}",
                "segments:": [{"str": "hello "}, {"dsl": "foo", "json": "#foo"}],
            },
        )

        validate_events(
            {
                (ProbePollerEvent.NEW_PROBES, frozenset({"probe3"})),
            }
        )

        sleep(0.5)

        # testing two things:
        #  1. after sleep 0.5 probe status should report 2 probes
        #  2. bad config raises ValueError
        with pytest.raises(ValueError):
            adapter(config_metadata("not-supported"), {})

        validate_events(
            {
                (ProbePollerEvent.STATUS_UPDATE, frozenset({"probe2", "probe3"})),
            }
        )

        # remove configuration
        adapter(config_metadata("metricProbe_probe2"), None)

        validate_events(
            {
                (ProbePollerEvent.DELETED_PROBES, frozenset({"probe2"})),
            }
        )

    finally:
        config.diagnostics_interval = old_interval


def test_log_probe_attributes_parsing():
    probe = probe_factory(
        {
            "id": "3d338829-21c4-4a8a-8a1a-71fbce995efa",
            "version": 0,
            "type": ProbeType.LOG_PROBE,
            "language": "python",
            "where": {
                "sourceFile": "foo.py",
                "lines": ["57"],
            },
            "tags": ["env:staging", "version:v12417452-d2552757"],
            "template": "{weekID} {idea}",
            "segments": [
                {"dsl": "weekID", "json": {"eq": [{"ref": "weekID"}, 1]}},
                {"str": " "},
                {"dsl": "idea", "json": {"ref": "idea"}},
            ],
            "captureSnapshot": False,
            "capture": {"maxReferenceDepth": 42, "maxLength": 43},
            "sampling": {"snapshotsPerSecond": 5000},
        }
    )

    assert isinstance(probe, LogProbeMixin)

    assert probe.limits.max_level == 42
    assert probe.limits.max_len == 43


def test_parse_log_probe_with_rate():
    probe = probe_factory(
        {
            "id": "3d338829-21c4-4a8a-8a1a-71fbce995efa",
            "version": 0,
            "type": ProbeType.LOG_PROBE,
            "tags": ["foo:bar"],
            "where": {"sourceFile": "tests/submod/stuff.p", "lines": ["36"]},
            "template": "hello {#foo}",
            "segments:": [{"str": "hello "}, {"dsl": "foo", "json": "#foo"}],
            "sampling": {"snapshotsPerSecond": 1337},
        }
    )

    assert probe.rate == 1337


def test_parse_log_probe_default_rates():
    probe = probe_factory(
        {
            "id": "3d338829-21c4-4a8a-8a1a-71fbce995efa",
            "version": 0,
            "type": ProbeType.LOG_PROBE,
            "tags": ["foo:bar"],
            "where": {"sourceFile": "tests/submod/stuff.p", "lines": ["36"]},
            "template": "hello {#foo}",
            "segments:": [{"str": "hello "}, {"dsl": "foo", "json": "#foo"}],
            "captureSnapshot": True,
        }
    )

    assert probe.rate == DEFAULT_SNAPSHOT_PROBE_RATE

    probe = probe_factory(
        {
            "id": "3d338829-21c4-4a8a-8a1a-71fbce995efa",
            "version": 0,
            "type": ProbeType.LOG_PROBE,
            "tags": ["foo:bar"],
            "where": {"sourceFile": "tests/submod/stuff.p", "lines": ["36"]},
            "template": "hello {#foo}",
            "segments:": [{"str": "hello "}, {"dsl": "foo", "json": "#foo"}],
            "captureSnapshot": False,
        }
    )

    assert probe.rate == DEFAULT_PROBE_RATE



# -*- encoding: utf-8 -*-
import os
import re
import subprocess
import sys
import time

import pytest

from . import utils


# gunicorn is not available on Windows
if sys.platform == "win32":
    pytestmark = pytest.mark.skip

TESTING_GEVENT = os.getenv("DD_PROFILE_TEST_GEVENT", False)


def _run_gunicorn(*args):
    cmd = (
        ["ddtrace-run", "gunicorn", "--bind", "127.0.0.1:7643", "--chdir", os.path.dirname(__file__)]
        + list(args)
        + ["gunicorn-app:app"]
    )
    return subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)


@pytest.fixture
def gunicorn(monkeypatch):
    # Do not ignore profiler so we have samples in the output pprof
    monkeypatch.setenv("DD_PROFILING_IGNORE_PROFILER", "0")
    monkeypatch.setenv("DD_PROFILING_ENABLED", "1")

    yield _run_gunicorn


def _get_worker_pids(stdout, num_worker, num_app_started=1):
    # type: (...) -> list[int]
    worker_pids = []

    while len(worker_pids) != num_worker:
        line = stdout.readline()
        if line == b"":
            raise RuntimeError("Boot failure")

        m = re.search(r"Booting worker with pid: (\d+)", line.decode())
        if m:
            worker_pids.append(int(m.group(1)))

    return worker_pids


def _test_gunicorn(gunicorn, tmp_path, monkeypatch, *args):
    # type: (...) -> None
    filename = str(tmp_path / "gunicorn.pprof")
    monkeypatch.setenv("DD_PROFILING_OUTPUT_PPROF", filename)
    proc = gunicorn(*args)
    worker_pids = _get_worker_pids(proc.stdout, 1)
    assert len(worker_pids) == 1
    time.sleep(3)
    proc.terminate()
    assert proc.wait() == 0
    utils.check_pprof_file("%s.%d.1" % (filename, proc.pid))
    for pid in worker_pids:
        utils.check_pprof_file("%s.%d.1" % (filename, pid))


@pytest.mark.skipif(TESTING_GEVENT, reason="Only testing gevent worker class")
def test_gunicorn(gunicorn, tmp_path, monkeypatch):
    # type: (...) -> None
    _test_gunicorn(gunicorn, tmp_path, monkeypatch)



import grpc

from ddtrace.internal.compat import ensure_pep562
from ddtrace.vendor.debtcollector import deprecate


GRPC_PIN_MODULE_SERVER = grpc.Server
GRPC_PIN_MODULE_CLIENT = grpc.Channel
GRPC_METHOD_PATH_KEY = "grpc.method.path"
GRPC_METHOD_PACKAGE_KEY = "grpc.method.package"
GRPC_METHOD_SERVICE_KEY = "grpc.method.service"
GRPC_METHOD_NAME_KEY = "grpc.method.name"
GRPC_METHOD_KIND_KEY = "grpc.method.kind"
GRPC_STATUS_CODE_KEY = "grpc.status.code"
GRPC_REQUEST_METADATA_PREFIX_KEY = "grpc.request.metadata."
GRPC_RESPONSE_METADATA_PREFIX_KEY = "grpc.response.metadata."
GRPC_HOST_KEY = "grpc.host"
GRPC_SPAN_KIND_KEY = "span.kind"
GRPC_SPAN_KIND_VALUE_CLIENT = "client"
GRPC_SPAN_KIND_VALUE_SERVER = "server"
GRPC_METHOD_KIND_UNARY = "unary"
GRPC_METHOD_KIND_CLIENT_STREAMING = "client_streaming"
GRPC_METHOD_KIND_SERVER_STREAMING = "server_streaming"
GRPC_METHOD_KIND_BIDI_STREAMING = "bidi_streaming"
GRPC_SERVICE_SERVER = "grpc-server"
GRPC_AIO_SERVICE_SERVER = "grpc-aio-server"
GRPC_SERVICE_CLIENT = "grpc-client"
GRPC_AIO_SERVICE_CLIENT = "grpc-aio-client"


def __getattr__(name):
    if name == "GRPC_PORT_KEY":
        deprecate(
            ("%s.%s is deprecated" % (__name__, name)),
            postfix=". Use ddtrace.ext.net.TARGET_PORT instead.",
            removal_version="2.0.0",
        )
        return "grpc.port"

    if name in globals():
        return globals()[name]

    raise AttributeError("'%s' has no attribute '%s'", __name__, name)


ensure_pep562(__name__)



import os
import typing as t
from warnings import warn

from envier import En


def parse_venv(value):
    # type: (str) -> t.Optional[str]
    try:
        return os.path.abspath(value) if value is not None else None
    except TypeError:
        warn(
            "No virtual environment detected. Running without a virtual environment active might "
            "cause exploration tests to instrument more than intended."
        )


class ExplorationConfig(En):
    venv = En.v(
        t.Optional[str],
        "virtual_env",
        parser=parse_venv,
        default=None,
    )

    encode = En.v(
        bool,
        "dd.debugger.expl.encode",
        default=True,
        help="Whether to encode the snapshots",
    )

    status_messages = En.v(
        bool,
        "dd.debugger.expl.status_messages",
        default=False,
        help="Whether to print exploration debugger status messages",
    )

    include = En.v(
        list,
        "dd.debugger.expl.include",
        parser=lambda v: [path.split(".") for path in v.split(",")],
        default=[],
        help="List of module paths to include in the exploration",
    )

    elusive = En.v(
        bool,
        "dd.debugger.expl.elusive",
        default=False,
        help="Whether to include elusive modules in the exploration",
    )

    class ProfilerConfig(En):
        __item__ = "profiler"
        __prefix__ = "dd.debugger.expl.profiler"

        enabled = En.v(
            bool,
            "enabled",
            default=True,
            help="Whether to enable the exploration deterministic profiler",
        )

    class CoverageConfig(En):
        __item__ = "coverage"
        __prefix__ = "dd.debugger.expl.coverage"

        enabled = En.v(
            bool,
            "enabled",
            default=True,
            help="Whether to enable the exploration line coverage",
        )

        delete_probes = En.v(
            bool,
            "delete_line_probes",
            default=False,
            help="Whether to delete line probes after they are triggered",
        )


config = ExplorationConfig()



from ddtrace.internal.compat import ensure_pep562
from ddtrace.vendor.debtcollector import deprecate


SERVICE = "mongodb"
COLLECTION = "mongodb.collection"
DB = "mongodb.db"
ROWS = "mongodb.rows"
QUERY = "mongodb.query"


def __getattr__(name):
    if name == "ROWS":
        deprecate(
            ("%s.%s is deprecated" % (__name__, name)),
            postfix=". Use ddtrace.ext.db.ROWCOUNT instead.",
            removal_version="2.0.0",
        )
        return "mongodb.rows"

    if name in globals():
        return globals()[name]

    raise AttributeError("'%s' has no attribute '%s'", __name__, name)


ensure_pep562(__name__)



import errno
import json
import os
import os.path
from typing import Set
from typing import TYPE_CHECKING

import attr
from six import ensure_binary

from ddtrace.appsec import _asm_request_context
from ddtrace.appsec._constants import APPSEC
from ddtrace.appsec._constants import DEFAULT
from ddtrace.appsec._constants import SPAN_DATA_NAMES
from ddtrace.appsec._constants import WAF_ACTIONS
from ddtrace.appsec._constants import WAF_CONTEXT_NAMES
from ddtrace.appsec._constants import WAF_DATA_NAMES
from ddtrace.appsec.ddwaf import DDWaf
from ddtrace.appsec.ddwaf import version
from ddtrace.constants import MANUAL_KEEP_KEY
from ddtrace.constants import ORIGIN_KEY
from ddtrace.constants import RUNTIME_FAMILY
from ddtrace.contrib import trace_utils
from ddtrace.contrib.trace_utils import _normalize_tag_name
from ddtrace.ext import SpanTypes
from ddtrace.internal import _context
from ddtrace.internal.logger import get_logger
from ddtrace.internal.processor import SpanProcessor
from ddtrace.internal.rate_limiter import RateLimiter
from ddtrace.internal.telemetry import telemetry_metrics_writer
from ddtrace.internal.telemetry.constants import TELEMETRY_NAMESPACE_TAG_APPSEC


try:
    from json.decoder import JSONDecodeError
except ImportError:
    # handling python 2.X import error
    JSONDecodeError = ValueError  # type: ignore

if TYPE_CHECKING:  # pragma: no cover
    from typing import Any
    from typing import Dict
    from typing import List
    from typing import Tuple
    from typing import Union

    from ddtrace.appsec.ddwaf.ddwaf_types import ddwaf_context_capsule
    from ddtrace.span import Span


log = get_logger(__name__)


def _transform_headers(data):
    # type: (Union[Dict[str, str], List[Tuple[str, str]]]) -> Dict[str, Union[str, List[str]]]
    normalized = {}  # type: Dict[str, Union[str, List[str]]]
    headers = data if isinstance(data, list) else data.items()
    for header, value in headers:
        header = header.lower()
        if header in ("cookie", "set-cookie"):
            continue
        if header in normalized:  # if a header with the same lowercase name already exists, let's make it an array
            existing = normalized[header]
            if isinstance(existing, list):
                existing.append(value)
            else:
                normalized[header] = [existing, value]
        else:
            normalized[header] = value
    return normalized


def get_rules():
    # type: () -> str
    return os.getenv("DD_APPSEC_RULES", default=DEFAULT.RULES)


def get_appsec_obfuscation_parameter_key_regexp():
    # type: () -> bytes
    return ensure_binary(
        os.getenv("DD_APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP", DEFAULT.APPSEC_OBFUSCATION_PARAMETER_KEY_REGEXP)
    )


def get_appsec_obfuscation_parameter_value_regexp():
    # type: () -> bytes
    return ensure_binary(
        os.getenv("DD_APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP", DEFAULT.APPSEC_OBFUSCATION_PARAMETER_VALUE_REGEXP)
    )


_COLLECTED_REQUEST_HEADERS = {
    "accept",
    "accept-encoding",
    "accept-language",
    "cf-connecting-ip",
    "cf-connecting-ipv6",
    "content-encoding",
    "content-language",
    "content-length",
    "content-type",
    "fastly-client-ip",
    "forwarded",
    "forwarded-for",
    "host",
    "true-client-ip",
    "user-agent",
    "via",
    "x-client-ip",
    "x-cluster-client-ip",
    "x-forwarded",
    "x-forwarded-for",
    "x-real-ip",
}


def _set_headers(span, headers, kind):
    # type: (Span, Any, str) -> None
    for k in headers:
        if isinstance(k, tuple):
            key, value = k
        else:
            key, value = k, headers[k]
        if key.lower() in _COLLECTED_REQUEST_HEADERS:
            # since the header value can be a list, use `set_tag()` to ensure it is converted to a string
            span.set_tag(_normalize_tag_name(kind, key), value)


def _get_rate_limiter():
    # type: () -> RateLimiter
    return RateLimiter(int(os.getenv("DD_APPSEC_TRACE_RATE_LIMIT", DEFAULT.TRACE_RATE_LIMIT)))


def _get_waf_timeout():
    # type: () -> int
    return int(os.getenv("DD_APPSEC_WAF_TIMEOUT", DEFAULT.WAF_TIMEOUT))


@attr.s(eq=False)
class AppSecSpanProcessor(SpanProcessor):
    rules = attr.ib(type=str, factory=get_rules)
    obfuscation_parameter_key_regexp = attr.ib(type=bytes, factory=get_appsec_obfuscation_parameter_key_regexp)
    obfuscation_parameter_value_regexp = attr.ib(type=bytes, factory=get_appsec_obfuscation_parameter_value_regexp)
    _ddwaf = attr.ib(type=DDWaf, default=None)
    _addresses_to_keep = attr.ib(type=Set[str], factory=set)
    _rate_limiter = attr.ib(type=RateLimiter, factory=_get_rate_limiter)
    _waf_timeout = attr.ib(type=int, factory=_get_waf_timeout)

    @property
    def enabled(self):
        return self._ddwaf is not None

    def __attrs_post_init__(self):
        # type: () -> None
        if self._ddwaf is None:
            try:
                with open(self.rules, "r") as f:
                    rules = json.load(f)
            except EnvironmentError as err:
                if err.errno == errno.ENOENT:
                    log.error(
                        "[DDAS-0001-03] ASM could not read the rule file %s. Reason: file does not exist", self.rules
                    )
                else:
                    # TODO: try to log reasons
                    log.error("[DDAS-0001-03] ASM could not read the rule file %s.", self.rules)
                raise
            except JSONDecodeError:
                log.error("[DDAS-0001-03] ASM could not read the rule file %s. Reason: invalid JSON file", self.rules)
                raise
            except Exception:
                # TODO: try to log reasons
                log.error("[DDAS-0001-03] ASM could not read the rule file %s.", self.rules)
                raise
            try:
                self._ddwaf = DDWaf(
                    rules, self.obfuscation_parameter_key_regexp, self.obfuscation_parameter_value_regexp
                )
            except ValueError:
                # Partial of DDAS-0005-00
                log.warning("[DDAS-0005-00] WAF initialization failed")
                raise
        for address in self._ddwaf.required_data:
            self._mark_needed(address)
        # we always need the request headers
        self._mark_needed(WAF_DATA_NAMES.REQUEST_HEADERS_NO_COOKIES)
        # we always need the response headers
        self._mark_needed(WAF_DATA_NAMES.RESPONSE_HEADERS_NO_COOKIES)

    def _update_rules(self, new_rules):
        # type: (Dict[str, Any]) -> bool
        result = False
        try:
            result = self._ddwaf.update_rules(new_rules)
        except TypeError:
            log.debug("Error updating ASM rules", exc_info=True)
        return result

    def _set_metrics(self):
        try:
            list_results, list_result_info, list_is_blocked = _asm_request_context.get_waf_results()
            if any((list_results, list_result_info, list_is_blocked)):
                is_blocked = any(list_is_blocked)
                is_triggered = any((result.data for result in list_results))
                has_info = any(list_result_info)

                tags = {
                    "waf_version": version(),
                    "lib_language": "python",
                    "rule_triggered": is_triggered,
                    "request_blocked": is_blocked,
                }

                if has_info:
                    for ddwaf_info in list_result_info:
                        if ddwaf_info.version:
                            tags["event_rules_version"] = ddwaf_info.version
                            telemetry_metrics_writer.add_count_metric(
                                TELEMETRY_NAMESPACE_TAG_APPSEC,
                                "event_rules.loaded",
                                float(ddwaf_info.loaded),
                                tags=tags,
                            )
                if list_results:
                    # runtime is the result in microseconds. Update to milliseconds
                    ddwaf_result_runtime = sum(float(ddwaf_result.runtime) for ddwaf_result in list_results)
                    ddwaf_result_total_runtime = sum(float(ddwaf_result.runtime) for ddwaf_result in list_results)
                    telemetry_metrics_writer.add_distribution_metric(
                        TELEMETRY_NAMESPACE_TAG_APPSEC,
                        "waf.duration",
                        float(ddwaf_result_runtime / 1e3),
                        tags=tags,
                    )
                    telemetry_metrics_writer.add_distribution_metric(
                        TELEMETRY_NAMESPACE_TAG_APPSEC,
                        "waf.duration_ext",
                        float(ddwaf_result_total_runtime / 1e3),
                        tags=tags,
                    )

                telemetry_metrics_writer.add_count_metric(
                    TELEMETRY_NAMESPACE_TAG_APPSEC,
                    "waf.requests",
                    1.0,
                    tags=tags,
                )
                # TODO: add log metric to report info.failed and info.errors
        except Exception:
            log.warning("Error reporting ASM metrics: %s", exc_info=True)
        finally:
            _asm_request_context.reset_waf_results()

    def on_span_start(self, span):
        # type: (Span) -> None

        if span.span_type != SpanTypes.WEB:
            return
        ctx = self._ddwaf._at_request_start()

        peer_ip = _asm_request_context.get_ip()
        headers = _asm_request_context.get_headers()
        headers_case_sensitive = _asm_request_context.get_headers_case_sensitive()

        span.set_metric(APPSEC.ENABLED, 1.0)
        span.set_tag_str(RUNTIME_FAMILY, "python")

        def waf_callable(custom_data=None):
            return self._waf_action(span._local_root or span, ctx, custom_data)

        _asm_request_context.set_waf_callback(waf_callable)

        if headers is not None:
            _context.set_items(
                {
                    SPAN_DATA_NAMES.REQUEST_HEADERS_NO_COOKIES: headers,
                    SPAN_DATA_NAMES.REQUEST_HEADERS_NO_COOKIES_CASE: headers_case_sensitive,
                },
                span=span,
            )
            if not peer_ip:
                return

            ip = trace_utils._get_request_header_client_ip(headers, peer_ip, headers_case_sensitive)
            # Save the IP and headers in the context so the retrieval can be skipped later
            _context.set_item("http.request.remote_ip", ip, span=span)
            if ip and self._is_needed(WAF_DATA_NAMES.REQUEST_HTTP_IP):
                log.debug("[DDAS-001-00] Executing ASM WAF for checking IP block")
                # _asm_request_context.call_callback()
                _asm_request_context.call_waf_callback({"REQUEST_HTTP_IP": None})

    def _waf_action(self, span, ctx, custom_data=None):
        # type: (Span, ddwaf_context_capsule, dict[str, Any] | None) -> None
        """
        Call the `WAF` with the given parameters. If `custom_data_names` is specified as
        a list of `(WAF_NAME, WAF_STR)` tuples specifying what values of the `WAF_DATA_NAMES`
        constant class will be checked. Else, it will check all the possible values
        from `WAF_DATA_NAMES`.

        If `custom_data_values` is specified, it must be a dictionary where the key is the
        `WAF_DATA_NAMES` key and the value the custom value. If not used, the values will
        be retrieved from the `_context`. This can be used when you don't want to store
        the value in the `_context` before checking the `WAF`.
        """

        if span.span_type != SpanTypes.WEB:
            return

        if _context.get_item(WAF_CONTEXT_NAMES.BLOCKED, span=span):
            return

        data = {}
        iter_data = [(key, WAF_DATA_NAMES[key]) for key in custom_data] if custom_data is not None else WAF_DATA_NAMES

        # type ignore because mypy seems to not detect that both results of the if
        # above can iter if not None
        for key, waf_name in iter_data:  # type: ignore[attr-defined]
            if self._is_needed(waf_name):
                if custom_data is not None and custom_data.get(key) is not None:
                    value = custom_data.get(key)
                else:
                    value = _context.get_item(SPAN_DATA_NAMES[key], span=span)

                if value:
                    data[waf_name] = _transform_headers(value) if key.endswith("HEADERS_NO_COOKIES") else value
                    log.debug("[action] WAF got value %s", SPAN_DATA_NAMES[key])
                else:
                    log.debug("[action] WAF missing value %s", SPAN_DATA_NAMES[key])

        log.debug("[DDAS-001-00] Executing ASM In-App WAF")
        waf_results = self._ddwaf.run(ctx, data, self._waf_timeout)
        if waf_results and waf_results.data:
            log.debug("[DDAS-011-00] ASM In-App WAF returned: %s", waf_results.data)

        blocked = WAF_ACTIONS.BLOCK in waf_results.actions
        _asm_request_context.set_waf_results(waf_results, self._ddwaf.info, blocked)
        if blocked:
            _context.set_item(WAF_CONTEXT_NAMES.BLOCKED, True, span=span)

        try:
            info = self._ddwaf.info
            if info.errors:
                span.set_tag_str(APPSEC.EVENT_RULE_ERRORS, json.dumps(info.errors))
            span.set_tag_str(APPSEC.EVENT_RULE_VERSION, info.version)
            span.set_tag_str(APPSEC.WAF_VERSION, version())

            def update_metric(name, value):
                old_value = span.get_metric(name)
                if old_value is None:
                    old_value = 0.0
                span.set_metric(name, value + old_value)

            span.set_metric(APPSEC.EVENT_RULE_LOADED, info.loaded)
            span.set_metric(APPSEC.EVENT_RULE_ERROR_COUNT, info.failed)
            if waf_results:
                update_metric(APPSEC.WAF_DURATION, waf_results.runtime)
                update_metric(APPSEC.WAF_DURATION_EXT, waf_results.total_runtime)
        except (JSONDecodeError, ValueError):
            log.warning("Error parsing data ASM In-App WAF metrics report %s", info.errors)
        except Exception:
            log.warning("Error executing ASM In-App WAF metrics report: %s", exc_info=True)

        if (waf_results and waf_results.data) or blocked:
            # We run the rate limiter only if there is an attack, its goal is to limit the number of collected asm
            # events
            allowed = self._rate_limiter.is_allowed(span.start_ns)
            if not allowed:
                # TODO: add metric collection to keep an eye (when it's name is clarified)
                return

            for id_tag, kind in [
                (SPAN_DATA_NAMES.REQUEST_HEADERS_NO_COOKIES, "request"),
                (SPAN_DATA_NAMES.RESPONSE_HEADERS_NO_COOKIES, "response"),
            ]:
                headers_req = _context.get_item(id_tag, span=span)
                if headers_req:
                    _set_headers(span, headers_req, kind=kind)

            if waf_results and waf_results.data:
                span.set_tag_str(APPSEC.JSON, '{"triggers":%s}' % (waf_results.data,))
            if blocked:
                span.set_tag(APPSEC.BLOCKED, "true")
                self._set_metrics()

            # Partial DDAS-011-00
            span.set_tag_str(APPSEC.EVENT, "true")

            remote_ip = _context.get_item(SPAN_DATA_NAMES.REQUEST_HTTP_IP, span=span)
            if remote_ip:
                # Note that if the ip collection is disabled by the env var
                # DD_TRACE_CLIENT_IP_HEADER_DISABLED actor.ip won't be sent
                span.set_tag_str("actor.ip", remote_ip)

            # Right now, we overwrite any value that could be already there. We need to reconsider when ASM/AppSec's
            # specs are updated.
            span.set_tag(MANUAL_KEEP_KEY)
            if span.get_tag(ORIGIN_KEY) is None:
                span.set_tag_str(ORIGIN_KEY, APPSEC.ORIGIN_VALUE)

    def _mark_needed(self, address):
        # type: (str) -> None
        self._addresses_to_keep.add(address)

    def _is_needed(self, address):
        # type: (str) -> bool
        return address in self._addresses_to_keep

    def on_span_finish(self, span):
        # type: (Span) -> None

        if span.span_type != SpanTypes.WEB:
            return
        # this call is only necessary for tests or frameworks that are not using blocking
        if span.get_tag(APPSEC.JSON) is None:
            log.debug("metrics waf call")
            _asm_request_context.call_waf_callback()
        self._set_metrics()
        self._ddwaf._at_request_end()
        # Force to set respond headers at the end
        headers_req = _context.get_item(SPAN_DATA_NAMES.RESPONSE_HEADERS_NO_COOKIES, span=span)
        if headers_req:
            _set_headers(span, headers_req, kind="response")



import itertools
import os
import time
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from ...internal import atexit
from ...internal import forksafe
from ...settings import _config as config
from ..agent import get_connection
from ..agent import get_trace_url
from ..compat import get_connection_response
from ..compat import httplib
from ..encoding import JSONEncoderV2
from ..logger import get_logger
from ..periodic import PeriodicService
from ..runtime import get_runtime_id
from ..service import ServiceStatus
from ..utils.formats import asbool
from ..utils.time import StopWatch
from ..utils.version import _pep440_to_semver
from .constants import TELEMETRY_METRIC_TYPE_COUNT
from .constants import TELEMETRY_METRIC_TYPE_DISTRIBUTIONS
from .constants import TELEMETRY_METRIC_TYPE_GAUGE
from .constants import TELEMETRY_METRIC_TYPE_RATE
from .constants import TELEMETRY_TYPE_DISTRIBUTION
from .constants import TELEMETRY_TYPE_GENERATE_METRICS
from .data import get_application
from .data import get_dependencies
from .data import get_host_info
from .metrics import MetricTagType
from .metrics import MetricType
from .metrics_namespaces import MetricNamespace
from .metrics_namespaces import NamespaceMetricType


log = get_logger(__name__)


def _get_heartbeat_interval_or_default():
    # type: () -> float
    return float(os.getenv("DD_TELEMETRY_HEARTBEAT_INTERVAL", default=60))


def _get_telemetry_metrics_interval_or_default():
    # type: () -> float
    return float(os.getenv("DD_TELEMETRY_METRICS_INTERVAL_SECONDS", default=10))


class _TelemetryClient:
    def __init__(self, endpoint):
        # type: (str) -> None
        self._agent_url = get_trace_url()
        self._endpoint = endpoint
        self._encoder = JSONEncoderV2()
        self._headers = {
            "Content-type": "application/json",
            "DD-Client-Library-Language": "python",
            "DD-Client-Library-Version": _pep440_to_semver(),
        }

    @property
    def url(self):
        return "%s/%s" % (self._agent_url, self._endpoint)

    def send_event(self, request):
        # type: (Dict) -> Optional[httplib.HTTPResponse]
        """Sends a telemetry request to the trace agent"""
        resp = None
        try:
            rb_json = self._encoder.encode(request)
            headers = self.get_headers(request)
            with StopWatch() as sw:
                conn = get_connection(self._agent_url)
                conn.request("POST", self._endpoint, rb_json, headers)
                resp = get_connection_response(conn)
            if resp.status < 300:
                log.debug("sent %d in %.5fs to %s. response: %s", len(rb_json), sw.elapsed(), self.url, resp.status)
            else:
                log.debug("failed to send telemetry to the Datadog Agent at %s. response: %s", self.url, resp.status)
        except Exception:
            log.debug("failed to send telemetry to the Datadog Agent at %s.", self.url, exc_info=True)
        finally:
            conn.close()
        return resp

    def get_headers(self, request):
        # type: (Dict) -> Dict
        """Get all telemetry api v1 request headers"""
        headers = self._headers.copy()
        headers["DD-Telemetry-Debug-Enabled"] = request["debug"]
        headers["DD-Telemetry-Request-Type"] = request["request_type"]
        headers["DD-Telemetry-API-Version"] = request["api_version"]
        headers["DD-Agent-Hostname"] = request["host"]["hostname"]
        if config.env:
            headers["DD-Agent-Env"] = config.env
        return headers


class TelemetryBase(PeriodicService):
    """
    Common features of Telemetry services
    """

    # telemetry endpoint uses events platform v2 api
    ENDPOINT_V2 = "telemetry/proxy/api/v2/apmtelemetry"
    # Counter representing the number of events sent by the writer. Here we are relying on the atomicity
    # of `itertools.count()` which is a CPython implementation detail. The sequence field in telemetry
    # payloads is only used in tests and is not required to process Telemetry events.
    _sequence = itertools.count(1)

    def __init__(self, interval):
        # type: (float) -> None
        super(TelemetryBase, self).__init__(interval=interval)

        # _enabled is None at startup, and is only set to true or false
        # after the config has been processed
        self._enabled = None  # type: Optional[bool]
        self._forked = False  # type: bool
        self._events_queue = []  # type: List[Dict]
        self._lock = forksafe.Lock()  # type: forksafe.ResetObject
        forksafe.register(self._fork_writer)

        # Debug flag that enables payload debug mode.
        self._debug = asbool(os.environ.get("DD_TELEMETRY_DEBUG", "false"))

        self._client = _TelemetryClient(self.ENDPOINT_V2)

    def add_event(self, payload, payload_type):
        # type: (Dict[str, Any], str) -> None
        """
        Adds a Telemetry event to the TelemetryWriter event buffer

        :param Dict payload: stores a formatted telemetry event
        :param str payload_type: The payload_type denotes the type of telmetery request.
            Payload types accepted by telemetry/proxy v1: app-started, app-closing, app-integrations-change
        """
        if self._enabled:
            event = {
                "tracer_time": int(time.time()),
                "runtime_id": get_runtime_id(),
                "api_version": "v1",
                "seq_id": next(self._sequence),
                "debug": self._debug,
                "application": get_application(config.service, config.version, config.env),
                "host": get_host_info(),
                "payload": payload,
                "request_type": payload_type,
            }
            self._events_queue.append(event)

    def enable(self):
        # type: () -> None
        """
        Enable the instrumentation telemetry collection service. If the service has already been
        activated before, this method does nothing. Use ``disable`` to turn off the telemetry collection service.
        """
        if self.status == ServiceStatus.RUNNING:
            return

        self._enabled = True
        self.start()

        atexit.register(self.stop)

    def disable(self):
        # type: () -> None
        """
        Disable the telemetry collection service and drop the existing integrations and events
        Once disabled, telemetry collection can be re-enabled by calling ``enable`` again.
        """
        with self._lock:
            self._enabled = False
        self.reset_queues()
        if self.status == ServiceStatus.STOPPED:
            return

        atexit.unregister(self.stop)

        self.stop()

    def reset_queues(self):
        # type: () -> None
        self._events_queue = []

    def _flush_events_queue(self):
        # type: () -> List[Dict]
        """Flushes and returns a list of all telemtery event"""
        with self._lock:
            events = self._events_queue
            self._events_queue = []
        return events

    def _fork_writer(self):
        # type: () -> None
        self._forked = True
        # Avoid sending duplicate events.
        # Queued events should be sent in the main process.
        self.reset_queues()

    def _restart_sequence(self):
        self._sequence = itertools.count(1)

    def _stop_service(self, *args, **kwargs):
        # type: (...) -> None
        super(TelemetryBase, self)._stop_service(*args, **kwargs)
        self.join()


class TelemetryMetricsWriter(TelemetryBase):
    """
    Submits Telemetry Metrics events to the datadog agent.
    """

    def __init__(self):
        # type: () -> None
        super(TelemetryMetricsWriter, self).__init__(interval=_get_telemetry_metrics_interval_or_default())
        self._namespace = MetricNamespace()

    def add_gauge_metric(self, namespace, name, value, tags={}):
        # type: (str,str, float, MetricTagType) -> None
        """
        Queues gauge metric
        """
        self._add_metric(TELEMETRY_METRIC_TYPE_GAUGE, namespace, name, value, tags)

    def add_rate_metric(self, namespace, name, value=1.0, tags={}):
        # type: (str,str, float, MetricTagType) -> None
        """
        Queues rate metric
        """
        self._add_metric(TELEMETRY_METRIC_TYPE_RATE, namespace, name, value, tags)

    def add_count_metric(self, namespace, name, value=1.0, tags={}):
        # type: (str,str, float, MetricTagType) -> None
        """
        Queues count metric
        """
        self._add_metric(TELEMETRY_METRIC_TYPE_COUNT, namespace, name, value, tags)

    def add_distribution_metric(self, namespace, name, value=1.0, tags={}):
        # type: (str,str, float, MetricTagType) -> None
        """
        Queues distributions metric
        """
        self._add_metric(TELEMETRY_METRIC_TYPE_DISTRIBUTIONS, namespace, name, value, tags)

    def _add_metric(self, metric_type, namespace, name, value=1.0, tags={}):
        # type: (MetricType, str,str, float, MetricTagType) -> None
        """
        Queues metric
        """
        if config._telemetry_metrics_enabled:
            with self._lock:
                self._namespace._add_metric(
                    metric_type, namespace, name, value, tags, interval=_get_heartbeat_interval_or_default()
                )

    def periodic(self):
        namespace_metrics = self._flush_namespace_metrics()
        if namespace_metrics:
            self._app_generate_metrics_event(namespace_metrics)

        telemetry_events = self._flush_events_queue()
        for telemetry_event in telemetry_events:
            self._client.send_event(telemetry_event)

    def _flush_namespace_metrics(self):
        # type () -> List[Metric]
        """Returns a list of all generated metrics and clears the namespace's list"""
        with self._lock:
            namespace_metrics = self._namespace.get()
            self._namespace._flush()
        return namespace_metrics

    def _app_generate_metrics_event(self, namespace_metrics):
        # type: (NamespaceMetricType) -> None
        for payload_type, namespaces in namespace_metrics.items():
            for namespace, metrics in namespaces.items():
                if metrics:
                    payload = {
                        "namespace": namespace,
                        "lib_language": "python",
                        "lib_version": _pep440_to_semver(),
                        "series": [m.to_dict() for m in metrics.values()],
                    }
                    log.debug("%s request payload, namespace %s", payload_type, namespace)
                    if payload_type == TELEMETRY_TYPE_DISTRIBUTION:
                        self.add_event(payload, TELEMETRY_TYPE_DISTRIBUTION)
                    elif payload_type == TELEMETRY_TYPE_GENERATE_METRICS:
                        self.add_event(payload, TELEMETRY_TYPE_GENERATE_METRICS)

    def on_shutdown(self):
        self.periodic()

    def reset_queues(self):
        # type: () -> None
        super(TelemetryMetricsWriter, self).reset_queues()
        self._namespace._flush()


class TelemetryWriter(TelemetryBase):
    """
    Submits Instrumentation Telemetry events to the datadog agent.
    Supports v2 of the instrumentation telemetry api
    """

    def __init__(self):
        # type: () -> None
        super(TelemetryWriter, self).__init__(interval=_get_heartbeat_interval_or_default())
        self._integrations_queue = []  # type: List[Dict]

    def add_integration(self, integration_name, auto_enabled):
        # type: (str, bool) -> None
        """
        Creates and queues the names and settings of a patched module

        :param str integration_name: name of patched module
        :param bool auto_enabled: True if module is enabled in _monkey.PATCH_MODULES
        """
        if self._enabled is None or self._enabled:
            # Integrations can be patched before the telemetry writer is enabled.
            integration = {
                "name": integration_name,
                "version": "",
                "enabled": True,
                "auto_enabled": auto_enabled,
                "compatible": True,
                "error": "",
            }
            self._integrations_queue.append(integration)

    def _app_started_event(self):
        # type: () -> None
        """Sent when TelemetryWriter is enabled or forks"""
        if self._forked:
            # app-started events should only be sent by the main process
            return
        payload = {
            "dependencies": get_dependencies(),
            "integrations": self._flush_integrations_queue(),
            "configurations": [],
        }
        self.add_event(payload, "app-started")

    def _app_heartbeat_event(self):
        # type: () -> None
        if self._forked:
            # TODO: Enable app-heartbeat on forks
            #   Since we only send app-started events in the main process
            #   any forked processes won't be able to access the list of
            #   dependencies for this app, and therefore app-heartbeat won't
            #   add much value today.
            return

        self.add_event({}, "app-heartbeat")

    def _app_closing_event(self):
        # type: () -> None
        """Adds a Telemetry event which notifies the agent that an application instance has terminated"""
        if self._forked:
            # app-closing event should only be sent by the main process
            return
        payload = {}  # type: Dict
        self.add_event(payload, "app-closing")

    def _app_integrations_changed_event(self, integrations):
        # type: (List[Dict]) -> None
        """Adds a Telemetry event which sends a list of configured integrations to the agent"""
        payload = {
            "integrations": integrations,
        }
        self.add_event(payload, "app-integrations-change")

    def periodic(self):
        integrations = self._flush_integrations_queue()
        if integrations:
            self._app_integrations_changed_event(integrations)

        if not self._events_queue:
            # Optimization: only queue heartbeat if no other events are queued
            self._app_heartbeat_event()

        telemetry_events = self._flush_events_queue()
        for telemetry_event in telemetry_events:
            self._client.send_event(telemetry_event)

    def _flush_integrations_queue(self):
        # type: () -> List[Dict]
        """Flushes and returns a list of all queued integrations"""
        with self._lock:
            integrations = self._integrations_queue
            self._integrations_queue = []
        return integrations

    def _start_service(self, *args, **kwargs):
        # type: (...) -> None
        self._app_started_event()
        return super(TelemetryBase, self)._start_service(*args, **kwargs)

    def on_shutdown(self):
        self._app_closing_event()
        self.periodic()

    def reset_queues(self):
        # type: () -> None
        super(TelemetryWriter, self).reset_queues()
        self._integrations_queue = []



import inspect
import os
import sys
from typing import TYPE_CHECKING


if TYPE_CHECKING:  # pragma: no cover
    from typing import Optional
    from typing import Text
    from typing import Tuple


FIRST_FRAME_NO_DDTRACE = 4

DD_TRACE_INSTALLED_PREFIX = os.sep + "ddtrace" + os.sep
TESTS_PREFIX = os.sep + "tests" + os.sep


def get_info_frame():
    # type: () -> Optional[Tuple[Text, int]]
    """Get the filename (path + filename) and line number of the original wrapped function to report it.

    CAVEAT: We should migrate this function to native code to improve the performance.
    """
    stack = inspect.stack()
    for frame in stack[FIRST_FRAME_NO_DDTRACE:]:
        if sys.version_info < (3, 0, 0):
            filename = frame[1]
            lineno = frame[2]
        else:
            filename = frame.filename
            lineno = frame.lineno

        if DD_TRACE_INSTALLED_PREFIX in filename and TESTS_PREFIX not in filename:
            continue

        return filename, lineno

    return None



import os
from typing import TYPE_CHECKING

from ddtrace import config
from ddtrace.appsec._constants import PRODUCTS
from ddtrace.appsec.utils import _appsec_rc_features_is_enabled
from ddtrace.constants import APPSEC_ENV
from ddtrace.internal.logger import get_logger
from ddtrace.internal.remoteconfig.client import RemoteConfigCallBack
from ddtrace.internal.remoteconfig.client import RemoteConfigCallBackAfterMerge
from ddtrace.internal.utils.formats import asbool


try:
    from json.decoder import JSONDecodeError
except ImportError:
    # handling python 2.X import error
    JSONDecodeError = ValueError  # type: ignore


if TYPE_CHECKING:  # pragma: no cover
    from typing import Any
    from typing import Dict

    try:
        from typing import Literal
    except ImportError:
        # Python < 3.8. The "type ignore" is to avoid a runtime check just to silence mypy.
        from typing_extensions import Literal  # type: ignore
    from typing import Mapping
    from typing import Optional
    from typing import Union

    from ddtrace import Tracer

log = get_logger(__name__)


def enable_appsec_rc(test_tracer=None):
    # type: (Optional[Tracer]) -> None
    # Tracer is a parameter for testing propose
    # Import tracer here to avoid a circular import
    if test_tracer is None:
        from ddtrace import tracer
    else:
        tracer = test_tracer

    asm_features_callback = RCAppSecFeaturesCallBack(tracer)
    asm_dd_callback = RCASMDDCallBack(tracer)
    asm_callback = RCAppSecCallBack(tracer)

    if _appsec_rc_features_is_enabled():
        from ddtrace.internal.remoteconfig import RemoteConfig

        RemoteConfig.register(PRODUCTS.ASM_FEATURES, asm_features_callback)

    if tracer._appsec_enabled:
        from ddtrace.internal.remoteconfig import RemoteConfig

        RemoteConfig.register(PRODUCTS.ASM_DATA, asm_callback)  # IP Blocking
        RemoteConfig.register(PRODUCTS.ASM, asm_callback)  # Exclusion Filters & Custom Rules
        RemoteConfig.register(PRODUCTS.ASM_DD, asm_dd_callback)  # DD Rules


def _add_rules_to_list(features, feature, message, ruleset):
    # type: (Mapping[str, Any], str, str, Dict[str, Any]) -> None
    rules = features.get(feature, None)
    if rules is not None:
        try:
            if ruleset.get(feature) is None:
                ruleset[feature] = []
            ruleset[feature] += rules
            log.debug("Reloading Appsec %s: %s", message, rules)
        except JSONDecodeError:
            log.error("ERROR Appsec %s: invalid JSON content from remote configuration", message)


def _appsec_rules_data(tracer, features):
    # type: (Tracer, Mapping[str, Any]) -> bool
    if features and tracer._appsec_processor:
        ruleset = {
            "rules": None,
            "rules_data": None,
            "exclusions": None,
            "rules_override": None,
        }  # type: dict[str, Optional[list[Any]]]
        _add_rules_to_list(features, "rules_data", "rules data", ruleset)
        _add_rules_to_list(features, "custom_rules", "custom rules", ruleset)
        _add_rules_to_list(features, "rules", "Datadog rules", ruleset)
        _add_rules_to_list(features, "exclusions", "exclusion filters", ruleset)
        _add_rules_to_list(features, "rules_override", "rules override", ruleset)
        return tracer._appsec_processor._update_rules({k: v for k, v in ruleset.items() if v is not None})

    return False


class RCASMDDCallBack(RemoteConfigCallBack):
    def __init__(self, tracer):
        # type: (Tracer) -> None
        self.tracer = tracer

    def __call__(self, metadata, features):
        if features is not None:
            _appsec_rules_data(self.tracer, features)


class RCAppSecFeaturesCallBack(RemoteConfigCallBack):
    def __init__(self, tracer):
        # type: (Tracer) -> None
        self.tracer = tracer

    def __call__(self, metadata, features):
        if features is not None:
            self._appsec_1click_activation(features)

    def _appsec_1click_activation(self, features):
        # type: (Union[Literal[False], Mapping[str, Any]]) -> None
        """This callback updates appsec enabled in tracer and config instances following this logic:
        ```
        | DD_APPSEC_ENABLED | RC Enabled | Result   |
        |-------------------|------------|----------|
        | <not set>         | <not set>  | Disabled |
        | <not set>         | false      | Disabled |
        | <not set>         | true       | Enabled  |
        | false             | <not set>  | Disabled |
        | true              | <not set>  | Enabled  |
        | false             | true       | Disabled |
        | true              | true       | Enabled  |
        ```
        """
        if APPSEC_ENV in os.environ:
            # no one click activation if var env is set
            rc_appsec_enabled = asbool(os.environ.get(APPSEC_ENV))
        elif features is False:
            rc_appsec_enabled = False
        else:
            rc_appsec_enabled = features.get("asm", {}).get("enabled")

        if rc_appsec_enabled is not None:
            from ddtrace.appsec._constants import PRODUCTS
            from ddtrace.internal.remoteconfig import RemoteConfig

            log.debug("Updating ASM Remote Configuration ASM_FEATURES: %s", rc_appsec_enabled)

            if rc_appsec_enabled:
                asm_dd_callback = RCASMDDCallBack(self.tracer)
                asm_callback = RCAppSecCallBack(self.tracer)
                RemoteConfig.register(PRODUCTS.ASM_DATA, asm_callback)  # IP Blocking
                RemoteConfig.register(PRODUCTS.ASM, asm_callback)  # Exclusion Filters & Custom Rules
                RemoteConfig.register(PRODUCTS.ASM_DD, asm_dd_callback)  # DD Rules
                if not self.tracer._appsec_enabled:
                    self.tracer.configure(appsec_enabled=True)
                else:
                    config._appsec_enabled = True

            else:
                RemoteConfig.unregister(PRODUCTS.ASM_DATA)
                RemoteConfig.unregister(PRODUCTS.ASM)
                RemoteConfig.unregister(PRODUCTS.ASM_DD)
                if self.tracer._appsec_enabled:
                    self.tracer.configure(appsec_enabled=False)
                else:
                    config._appsec_enabled = False


class RCAppSecCallBack(RemoteConfigCallBackAfterMerge):
    def __init__(self, tracer):
        # type: (Tracer) -> None
        super(RCAppSecCallBack, self).__init__()
        self.tracer = tracer

    def __call__(self, target, features):
        # type: (str, Any) -> None
        if features is not None:
            _appsec_rules_data(self.tracer, features)



import json
import logging

from flask import Response
from flask import request
import pytest

from ddtrace.appsec._constants import SPAN_DATA_NAMES
from ddtrace.appsec.trace_utils import block_request_if_user_blocked
from ddtrace.constants import APPSEC_JSON
from ddtrace.ext import http
from ddtrace.internal import _context
from ddtrace.internal import constants
from ddtrace.internal.compat import six
from ddtrace.internal.compat import urlencode
from tests.appsec.test_processor import RULES_GOOD_PATH
from tests.appsec.test_processor import RULES_SRB
from tests.appsec.test_processor import RULES_SRB_METHOD
from tests.appsec.test_processor import RULES_SRB_RESPONSE
from tests.appsec.test_processor import _ALLOWED_IP
from tests.appsec.test_processor import _BLOCKED_IP
from tests.contrib.flask import BaseFlaskTestCase
from tests.utils import override_env
from tests.utils import override_global_config


_BLOCKED_USER = "123456"
_ALLOWED_USER = "111111"


class FlaskAppSecTestCase(BaseFlaskTestCase):
    @pytest.fixture(autouse=True)
    def inject_fixtures(self, caplog):
        self._caplog = caplog

    def _aux_appsec_prepare_tracer(self, appsec_enabled=True):
        self.tracer._appsec_enabled = appsec_enabled
        # Hack: need to pass an argument to configure so that the processors are recreated
        self.tracer.configure(api_version="v0.4")

    def test_flask_simple_attack(self):
        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/.git?q=1")
            assert resp.status_code == 404
            # Read response data from the test client to close flask.request and flask.response spans
            assert resp.data is not None
            root_span = self.pop_spans()[0]

            appsec_json = root_span.get_tag(APPSEC_JSON)
            assert "triggers" in json.loads(appsec_json if appsec_json else "{}")
            assert _context.get_item("http.request.uri", span=root_span) == "http://localhost/.git?q=1"
            query = dict(_context.get_item("http.request.query", span=root_span))
            assert query == {"q": "1"} or query == {"q": ["1"]}

    def test_flask_path_params(self):
        @self.app.route("/params/<item>")
        def dynamic_url(item):
            return item

        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/params/attack")
            assert resp.status_code == 200
            # Read response data from the test client to close flask.request and flask.response spans
            assert resp.data is not None
            root_span = self.pop_spans()[0]

            flask_args = root_span.get_tag("flask.view_args.item")
            assert flask_args == "attack"

            path_params = _context.get_item("http.request.path_params", span=root_span)
            assert path_params == {"item": "attack"}

    def test_flask_path_params_attack(self):
        @self.app.route("/params/<item>")
        def dynamic_url(item):
            return item

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/params/w00tw00t.at.isc.sans.dfind")
            assert resp.status_code == 200

            root_span = self.pop_spans()[0]

            appsec_json = root_span.get_tag(APPSEC_JSON)
            assert "triggers" in json.loads(appsec_json if appsec_json else "{}")

            query = dict(_context.get_item("http.request.path_params", span=root_span))
            assert query == {"item": "w00tw00t.at.isc.sans.dfind"}

    def test_flask_querystrings(self):
        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            self.client.get("/?a=1&b&c=d")
            root_span = self.pop_spans()[0]
            query = dict(_context.get_item("http.request.query", span=root_span))
            assert query == {"a": "1", "b": "", "c": "d"} or query == {"a": ["1"], "b": [""], "c": ["d"]}
            self.client.get("/")
            root_span = self.pop_spans()[0]
            assert len(_context.get_item("http.request.query", span=root_span)) == 0

    def test_flask_cookie_sql_injection(self):
        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
            self._aux_appsec_prepare_tracer()
            self.client.set_cookie("localhost", "attack", "1' or '1' = '1'")
            resp = self.client.get("/")
            assert resp.status_code == 404
            root_span = self.pop_spans()[0]

            appsec_json = root_span.get_tag(APPSEC_JSON)
            assert "triggers" in json.loads(appsec_json if appsec_json else "{}")
            assert _context.get_item("http.request.cookies", span=root_span)["attack"] == "1' or '1' = '1'"
            query = dict(_context.get_item("http.request.cookies", span=root_span))
            assert query == {"attack": "1' or '1' = '1'"} or query == {"attack": ["1' or '1' = '1'"]}

    def test_flask_cookie(self):
        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            self.client.set_cookie("localhost", "testingcookie_key", "testingcookie_value")
            resp = self.client.get("/")
            assert resp.status_code == 404
            # Read response data from the test client to close flask.request and flask.response spans
            assert resp.data is not None
            root_span = self.pop_spans()[0]

            assert root_span.get_tag(APPSEC_JSON) is None
            assert (
                _context.get_item("http.request.cookies", span=root_span)["testingcookie_key"] == "testingcookie_value"
            )
            query = dict(_context.get_item("http.request.cookies", span=root_span))
            assert query == {"testingcookie_key": "testingcookie_value"} or query == {
                "testingcookie_key": ["testingcookie_value"]
            }

    def test_flask_useragent(self):
        self.client.get("/", headers={"User-Agent": "test/1.2.3"})
        root_span = self.pop_spans()[0]
        assert root_span.get_tag(http.USER_AGENT) == "test/1.2.3"

    def test_flask_client_ip_header_set_by_env_var_valid(self):
        with override_global_config(dict(_appsec_enabled=True, client_ip_header="X-Use-This")):
            self.client.get("/?a=1&b&c=d", headers={"HTTP_X_CLIENT_IP": "8.8.8.8", "X-Use-This": "4.4.4.4"})
            spans = self.pop_spans()
            root_span = spans[0]
            assert root_span.get_tag(http.CLIENT_IP) == "4.4.4.4"

    def test_flask_body_urlencoded(self):
        @self.app.route("/body", methods=["GET", "POST", "DELETE"])
        def body():
            data = dict(request.form)
            return str(data), 200

        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            data = {"mytestingbody_key": "mytestingbody_value"}
            payload = urlencode(data)

            self.client.post("/body", data=payload, content_type="application/x-www-form-urlencoded")

            root_span = self.pop_spans()[0]
            query = dict(_context.get_item("http.request.body", span=root_span))

            assert root_span.get_tag(APPSEC_JSON) is None
            assert query == {"mytestingbody_key": "mytestingbody_value"}

    def test_flask_body_urlencoded_appsec_disabled_then_no_body(self):
        with override_global_config(dict(_appsec_enabled=False)):
            self._aux_appsec_prepare_tracer()
            payload = urlencode({"mytestingbody_key": "mytestingbody_value"})
            self.client.post("/", data=payload, content_type="application/x-www-form-urlencoded")
            root_span = self.pop_spans()[0]

            assert not _context.get_item("http.request.body", span=root_span)

    def test_flask_request_body_urlencoded_attack(self):
        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            payload = urlencode({"attack": "1' or '1' = '1'"})
            self.client.post("/", data=payload, content_type="application/x-www-form-urlencoded")
            root_span = self.pop_spans()[0]
            query = dict(_context.get_item("http.request.body", span=root_span))
            assert "triggers" in json.loads(root_span.get_tag(APPSEC_JSON))
            assert query == {"attack": "1' or '1' = '1'"}

    def test_flask_body_json(self):
        @self.app.route("/body", methods=["GET", "POST", "DELETE"])
        def body():
            data = request.get_json()
            return str(data), 200

        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            payload = {"mytestingbody_key": "mytestingbody_value"}

            self.client.post("/body", json=payload, content_type="application/json")

            root_span = self.pop_spans()[0]
            query = dict(_context.get_item("http.request.body", span=root_span))

            assert root_span.get_tag(APPSEC_JSON) is None
            assert query == {"mytestingbody_key": "mytestingbody_value"}

    def test_flask_body_json_attack(self):
        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            payload = {"attack": "1' or '1' = '1'"}
            self.client.post("/", json=payload, content_type="application/json")
            root_span = self.pop_spans()[0]
            query = dict(_context.get_item("http.request.body", span=root_span))
            assert "triggers" in json.loads(root_span.get_tag(APPSEC_JSON))
            assert query == {"attack": "1' or '1' = '1'"}

    def test_flask_body_xml(self):
        @self.app.route("/body", methods=["GET", "POST", "DELETE"])
        def body():
            data = request.data
            return data, 200

        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            payload = "<mytestingbody_key>mytestingbody_value</mytestingbody_key>"
            response = self.client.post("/body", data=payload, content_type="application/xml")
            assert response.status_code == 200
            assert response.data == b"<mytestingbody_key>mytestingbody_value</mytestingbody_key>"

            root_span = self.pop_spans()[0]
            query = dict(_context.get_item("http.request.body", span=root_span))

            assert root_span.get_tag(APPSEC_JSON) is None
            assert query == {"mytestingbody_key": "mytestingbody_value"}

    def test_flask_body_xml_attack(self):
        with override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            payload = "<attack>1' or '1' = '1'</attack>"
            self.client.post("/", data=payload, content_type="application/xml")
            root_span = self.pop_spans()[0]
            query = dict(_context.get_item("http.request.body", span=root_span))

            assert "triggers" in json.loads(root_span.get_tag(APPSEC_JSON))
            assert query == {"attack": "1' or '1' = '1'"}

    def test_flask_body_json_empty_body_logs_warning(self):
        with self._caplog.at_level(logging.DEBUG), override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            self.client.post("/", data="", content_type="application/json")
            assert "Failed to parse werkzeug request body" in self._caplog.text

    def test_flask_body_json_bad_logs_warning(self):
        with self._caplog.at_level(logging.DEBUG), override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            self.client.post("/", data="not valid json", content_type="application/json")
            assert "Failed to parse werkzeug request body" in self._caplog.text

    def test_flask_body_xml_bad_logs_warning(self):
        with self._caplog.at_level(logging.DEBUG), override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            self.client.post("/", data="bad xml", content_type="application/xml")
            assert "Failed to parse werkzeug request body" in self._caplog.text

    def test_flask_body_xml_empty_logs_warning(self):
        with self._caplog.at_level(logging.DEBUG), override_global_config(dict(_appsec_enabled=True)):
            self._aux_appsec_prepare_tracer()
            self.client.post("/", data="", content_type="application/xml")
            assert "Failed to parse werkzeug request body" in self._caplog.text

    def test_flask_ipblock_nomatch_200_json(self):
        @self.app.route("/")
        def route():
            return "OK", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/", headers={"X-Real-Ip": _ALLOWED_IP})
            root_span = self.pop_spans()[0]
            assert resp.status_code == 200
            assert not _context.get_item("http.request.blocked", span=root_span)

    def test_flask_ipblock_match_403_json(self):
        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/foobar", headers={"X-Real-Ip": _BLOCKED_IP})
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            assert root_span.get_tag(http.STATUS_CODE) == "403"
            assert root_span.get_tag(http.URL) == "http://localhost/foobar"
            assert root_span.get_tag(http.METHOD) == "GET"
            assert root_span.get_tag(http.USER_AGENT).startswith("werkzeug/")
            assert root_span.get_tag(SPAN_DATA_NAMES.RESPONSE_HEADERS_NO_COOKIES + ".content-type") == "text/json"

    def test_flask_ipblock_manually_json(self):
        # Most tests of flask blocking are in the test_flask_snapshot, this just
        # test a manual call to the blocking callable stored in _asm_request_context
        @self.app.route("/block")
        def test_route():
            from ddtrace.appsec._asm_request_context import block_request

            return block_request()

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/block", headers={"X-REAL-IP": _ALLOWED_IP})
            # Should not block by IP but since the route is calling block_request it will be blocked
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                # not all flask versions have r.text
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON

            root_span = self.pop_spans()[0]
            assert root_span.get_tag(http.STATUS_CODE) == "403"
            assert root_span.get_tag(http.URL) == "http://localhost/block"
            assert root_span.get_tag(http.METHOD) == "GET"
            assert root_span.get_tag(http.USER_AGENT).startswith("werkzeug/")
            assert root_span.get_tag(SPAN_DATA_NAMES.RESPONSE_HEADERS_NO_COOKIES + ".content-type") == "text/json"

    def test_flask_userblock_json(self):
        @self.app.route("/checkuser/<user_id>")
        def test_route(user_id):
            from ddtrace import tracer

            block_request_if_user_blocked(tracer, user_id)
            return "Ok", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/checkuser/%s" % _BLOCKED_USER)
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                # not all flask versions have r.text
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON

            root_span = self.pop_spans()[0]
            assert root_span.get_tag(http.STATUS_CODE) == "403"
            assert root_span.get_tag(http.URL) == "http://localhost/checkuser/%s" % _BLOCKED_USER
            assert root_span.get_tag(http.METHOD) == "GET"
            assert root_span.get_tag(http.USER_AGENT).startswith("werkzeug/")
            assert root_span.get_tag(SPAN_DATA_NAMES.RESPONSE_HEADERS_NO_COOKIES + ".content-type") == "text/json"

            resp = self.client.get("/checkuser/%s" % _BLOCKED_USER, headers={"Accept": "text/html"})
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                # not all flask versions have r.text
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_HTML

            resp = self.client.get("/checkuser/%s" % _ALLOWED_USER, headers={"Accept": "text/html"})
            assert resp.status_code == 200

    def test_request_suspicious_request_block_match_query_value(self):
        @self.app.route("/")
        def test_route():
            return "Ok", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB)):
            self._aux_appsec_prepare_tracer()

            resp = self.client.get("/index.html?toto=xtrace")
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-001"]
            assert root_span.get_tag(http.STATUS_CODE) == "403"
            assert root_span.get_tag(http.URL) == "http://localhost/index.html?toto=xtrace"
            assert root_span.get_tag(http.METHOD) == "GET"
            assert root_span.get_tag(http.USER_AGENT).startswith("werkzeug/")
            assert root_span.get_tag(SPAN_DATA_NAMES.RESPONSE_HEADERS_NO_COOKIES + ".content-type") == "text/json"

    def test_request_suspicious_request_block_match_uri(self):
        @self.app.route("/")
        def test_route():
            return "Ok", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB)):
            self._aux_appsec_prepare_tracer()

            resp = self.client.get("/.git")
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-002"]
            assert root_span.get_tag(http.STATUS_CODE) == "403"
            assert root_span.get_tag(http.URL) == "http://localhost/.git"
            assert root_span.get_tag(http.METHOD) == "GET"
            assert root_span.get_tag(http.USER_AGENT).startswith("werkzeug/")
            assert root_span.get_tag(SPAN_DATA_NAMES.RESPONSE_HEADERS_NO_COOKIES + ".content-type") == "text/json"

    def test_request_suspicious_request_block_match_body(self):
        @self.app.route("/")
        def test_route():
            return "Ok", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.post(
                "/index.html",
                data='{"key": "yqrweytqwreasldhkuqwgervflnmlnli"}',
                content_type="application/json",
            )
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-003"]

    def test_request_suspicious_request_block_match_header(self):
        @self.app.route("/")
        def test_route():
            return "Ok", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB)):
            self._aux_appsec_prepare_tracer()

            resp = self.client.get("/", headers={"User-Agent": "01972498723465"})
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-004"]

    def test_request_suspicious_request_block_match_response_code(self):
        @self.app.route("/")
        def test_route():
            return "Ok", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB_RESPONSE)):
            self._aux_appsec_prepare_tracer()

            resp = self.client.get("/do_not_exist.php")
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-005"]

    def test_request_suspicious_request_block_match_method(self):
        @self.app.route("/")
        def test_route():
            return "Ok", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB_METHOD)):
            self._aux_appsec_prepare_tracer()

            resp = self.client.get("/do_not_exist.php")
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-006"]

    def test_request_suspicious_request_block_match_cookies(self):
        @self.app.route("/")
        def test_route():
            return "Ok", 200

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB)):
            self._aux_appsec_prepare_tracer()
            self.client.set_cookie("localhost", "keyname", "jdfoSDGFkivRG_234")
            resp = self.client.get("/")
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-008"]

    def test_request_suspicious_request_block_match_path_params(self):
        @self.app.route("/params/<item>")
        def dynamic_url(item):
            return item

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/params/AiKfOeRcvG45")
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            flask_args = root_span.get_tag("flask.view_args.item")
            assert flask_args == "AiKfOeRcvG45"
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-007"]

    def test_request_suspicious_request_block_match_response_headers(self):
        @self.app.route("/response-header/")
        def specific_reponse():
            resp = Response("Foo bar baz", 200)
            resp.headers["Content-Disposition"] = 'attachment; filename="MagicKey_Al4h7iCFep9s1"'
            return resp

        with override_global_config(dict(_appsec_enabled=True)), override_env(dict(DD_APPSEC_RULES=RULES_SRB)):
            self._aux_appsec_prepare_tracer()
            resp = self.client.get("/response-header/")
            assert resp.status_code == 403
            if hasattr(resp, "text"):
                assert resp.text == constants.APPSEC_BLOCKED_RESPONSE_JSON
            else:
                assert resp.data == six.ensure_binary(constants.APPSEC_BLOCKED_RESPONSE_JSON)
            root_span = self.pop_spans()[0]
            loaded = json.loads(root_span.get_tag(APPSEC_JSON))
            assert [t["rule"]["id"] for t in loaded["triggers"]] == ["tst-037-009"]



import os

from ddtrace.internal import atexit
from ddtrace.internal import forksafe
from ddtrace.internal.logger import get_logger
from ddtrace.internal.remoteconfig.worker import RemoteConfigWorker
from ddtrace.internal.utils.formats import asbool


log = get_logger(__name__)


class RemoteConfig(object):
    _worker = None
    _worker_lock = forksafe.Lock()

    @classmethod
    def enable(cls):
        # type: () -> bool
        # TODO: this is only temporary. DD_REMOTE_CONFIGURATION_ENABLED variable will be deprecated
        rc_env_enabled = asbool(os.environ.get("DD_REMOTE_CONFIGURATION_ENABLED", "true"))
        if rc_env_enabled:
            with cls._worker_lock:
                if cls._worker is None:
                    cls._worker = RemoteConfigWorker()
                    cls._worker.start()

                    forksafe.register(cls._restart)
                    atexit.register(cls.disable)
            return True
        return False

    @classmethod
    def _restart(cls):
        cls.disable()
        cls.enable()

    @classmethod
    def register(cls, product, handler):
        try:
            # By enabling on registration we ensure we start the RCM client only
            # if there is at least one registered product.
            if cls.enable():
                cls._worker._client.register_product(product, handler)
        except Exception:
            log.warning("error starting the RCM client", exc_info=True)

    @classmethod
    def unregister(cls, product):
        try:
            cls._worker._client.unregister_product(product)
        except Exception:
            log.warning("error starting the RCM client", exc_info=True)

    @classmethod
    def disable(cls, join=False):
        # type: (bool) -> None
        with cls._worker_lock:
            if cls._worker is not None:
                cls._worker.stop()
                if join:
                    cls._worker.join()
                cls._worker = None

                forksafe.unregister(cls._restart)
                atexit.unregister(cls.disable)

    def __enter__(self):
        # type: () -> RemoteConfig
        self.enable()
        return self

    def __exit__(self, *args):
        # type: (...) -> None
        self.disable(join=True)



from typing import Optional
from typing import TYPE_CHECKING

from ddtrace.appsec import _asm_request_context
from ddtrace.contrib.trace_utils import set_user
from ddtrace.internal import _context


if TYPE_CHECKING:
    from ddtrace import Span
    from ddtrace import Tracer

from ddtrace import config
from ddtrace import constants
from ddtrace.appsec._constants import APPSEC
from ddtrace.appsec._constants import WAF_CONTEXT_NAMES
from ddtrace.ext import user
from ddtrace.internal.compat import six
from ddtrace.internal.logger import get_logger


log = get_logger(__name__)


def _track_user_login_common(tracer, user_id, success, metadata=None):
    # type: (Tracer, str, bool, Optional[dict]) -> Optional[Span]

    span = tracer.current_root_span()
    if span:
        success_str = "success" if success else "failure"
        span.set_tag_str("%s.%s.track" % (APPSEC.USER_LOGIN_EVENT_PREFIX, success_str), "true")
        if metadata is not None:
            for k, v in six.iteritems(metadata):
                span.set_tag_str("%s.%s.%s" % (APPSEC.USER_LOGIN_EVENT_PREFIX, success_str, k), str(v))
        span.set_tag_str(constants.MANUAL_KEEP_KEY, "true")
        return span
    else:
        log.warning(
            "No root span in the current execution. Skipping track_user_success_login tags. "
            "See https://docs.datadoghq.com/security_platform/application_security/setup_and_configure/"
            "?tab=set_user&code-lang=python for more information.",
        )
    return None


def track_user_login_success_event(
    tracer,
    user_id,
    metadata=None,
    name=None,
    email=None,
    scope=None,
    role=None,
    session_id=None,
    propagate=False,
):
    # type: (Tracer, str, Optional[dict], Optional[str], Optional[str], Optional[str], Optional[str], Optional[str], bool) -> None # noqa: E501
    """
    Add a new login success tracking event. The parameters after metadata (name, email,
    scope, role, session_id, propagate) will be passed to the `set_user` function that will be called
    by this one, see:
    https://docs.datadoghq.com/logs/log_configuration/attributes_naming_convention/#user-related-attributes
    https://docs.datadoghq.com/security_platform/application_security/setup_and_configure/?tab=set_tag&code-lang=python

    :param tracer: tracer instance to use
    :param user_id: a string with the UserId
    :param metadata: a dictionary with additional metadata information to be stored with the event
    """

    span = _track_user_login_common(tracer, user_id, True, metadata)
    if not span:
        return

    # usr.id will be set by set_user
    set_user(tracer, user_id, name, email, scope, role, session_id, propagate)


def track_user_login_failure_event(tracer, user_id, exists, metadata=None):
    # type: (Tracer, str, bool, Optional[dict]) -> None
    """
    Add a new login failure tracking event.
    :param tracer: tracer instance to use
    :param user_id: a string with the UserId if exists=True or the username if not
    :param exists: a boolean indicating if the user exists in the system
    :param metadata: a dictionary with additional metadata information to be stored with the event
    """

    span = _track_user_login_common(tracer, user_id, False, metadata)
    if not span:
        return

    span.set_tag_str("%s.failure.%s" % (APPSEC.USER_LOGIN_EVENT_PREFIX, user.ID), str(user_id))
    exists_str = "true" if exists else "false"
    span.set_tag_str("%s.failure.%s" % (APPSEC.USER_LOGIN_EVENT_PREFIX, user.EXISTS), exists_str)


def track_custom_event(tracer, event_name, metadata):
    # type: (Tracer, str, dict) -> None
    """
    Add a new custom tracking event.

    :param tracer: tracer instance to use
    :param event_name: the name of the custom event
    :param metadata: a dictionary with additional metadata information to be stored with the event
    """

    if not event_name:
        log.warning("Empty event name given to track_custom_event. Skipping setting tags.")
        return

    if not metadata:
        log.warning("Empty metadata given to track_custom_event. Skipping setting tags.")
        return

    span = tracer.current_root_span()
    if not span:
        log.warning(
            "No root span in the current execution. Skipping track_custom_event tags. "
            "See https://docs.datadoghq.com/security_platform/application_security"
            "/setup_and_configure/"
            "?tab=set_user&code-lang=python for more information.",
        )
        return

    span.set_tag_str("%s.%s.track" % (APPSEC.CUSTOM_EVENT_PREFIX, event_name), "true")

    for k, v in six.iteritems(metadata):
        span.set_tag_str("%s.%s.%s" % (APPSEC.CUSTOM_EVENT_PREFIX, event_name, k), str(v))
        span.set_tag_str(constants.MANUAL_KEEP_KEY, "true")


def should_block_user(tracer, userid):  # type: (Tracer, str) -> bool
    """
    Return true if the specified User ID should be blocked.

    :param tracer: tracer instance to use
    :param userid: the ID of the user as registered by `set_user`
    """

    if not config._appsec_enabled:
        log.warning(
            "One click blocking of user ids is disabled. To use this feature please enable "
            "Application Security Monitoring"
        )
        return False

    # Early check to avoid calling the WAF if the request is already blocked
    span = tracer.current_root_span()
    if _context.get_item(WAF_CONTEXT_NAMES.BLOCKED, span=span):
        return True

    _asm_request_context.call_waf_callback(custom_data={"REQUEST_USER_ID": str(userid)})
    return bool(_context.get_item(WAF_CONTEXT_NAMES.BLOCKED, span=span))


def block_request():  # type: () -> None
    """
    Block the current request and return a 403 Unauthorized response. If the response
    has already been started to be sent this could not work. The behaviour of this function
    could be different among frameworks, but it usually involves raising some kind of internal Exception,
    meaning that if you capture the exception the request blocking could not work.
    """
    if not config._appsec_enabled:
        log.warning("block_request() is disabled. To use this feature please enable" "Application Security Monitoring")
        return

    _asm_request_context.block_request()


def block_request_if_user_blocked(tracer, userid):  # type: (Tracer, str) -> None
    """
    Check if the specified User ID should be blocked and if positive
    block the current request using `block_request`.

    :param tracer: tracer instance to use
    :param userid: the ID of the user as registered by `set_user`
    """
    if not config._appsec_enabled:
        log.warning("should_block_user call requires ASM to be enabled")
        return

    if should_block_user(tracer, userid):
        span = tracer.current_root_span()
        if span:
            span.set_tag_str(user.ID, str(userid))
        _asm_request_context.block_request()



from os import environ
import random
import string

from flask import Flask


app = Flask(__name__)


@app.route("/")
def entry_point():
    if environ.get("DUPLICATE_TAGS_SCENARIO", None):
        from ddtrace import tracer

        span = tracer.current_span()
        if span:
            for _ in range(100):
                span.set_tag(_, "a" * 100)
    elif environ.get("UNIQUE_TAGS_SCENARIO", None):
        from ddtrace import tracer

        span = tracer.current_span()
        if span:
            for num in range(100):
                span.set_tag(str(num), "".join(random.choices(string.ascii_letters, k=100)))

    return "Hello World!"



import sys

import pytest

from ddtrace.internal.telemetry.constants import TELEMETRY_NAMESPACE_TAG_APPSEC
from ddtrace.internal.telemetry.constants import TELEMETRY_NAMESPACE_TAG_TRACER
from ddtrace.internal.telemetry.constants import TELEMETRY_TYPE_DISTRIBUTION
from ddtrace.internal.telemetry.constants import TELEMETRY_TYPE_GENERATE_METRICS
from ddtrace.internal.telemetry.metrics_namespaces import TelemetryTypeError
from ddtrace.internal.utils.version import _pep440_to_semver
from tests.telemetry.test_writer import _get_request_body
from tests.utils import override_global_config


def _assert_metric(
    test_agent,
    expected_series,
    namespace=TELEMETRY_NAMESPACE_TAG_TRACER,
    type_paypload=TELEMETRY_TYPE_GENERATE_METRICS,
    seq_id=1,
):
    test_agent.telemetry_writer.periodic()
    events = test_agent.get_events()

    assert len([event for event in events if event["request_type"] == type_paypload]) == seq_id

    payload = {
        "namespace": namespace,
        "lib_language": "python",
        "lib_version": _pep440_to_semver(),
        "series": expected_series,
    }
    assert events[0]["request_type"] == type_paypload

    assert events[0] == _get_request_body(payload, type_paypload, seq_id)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="mock.ANY doesn't works in py3.5 or lower")
def test_send_metric_flush_and_generate_metrics_series_is_restarted(test_agent_metrics_session, mock_time):
    """Check the queue of metrics is empty after run periodic method of PeriodicService"""
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric2", 1, {"a": "b"})
        expected_series = [
            {
                "common": True,
                "metric": "test-metric2",
                "points": [[1642544540, 1.0]],
                "tags": ["a:b"],
                "type": "count",
            },
        ]

        _assert_metric(test_agent_metrics_session, expected_series)

        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric2", 1, {"a": "b"})

        _assert_metric(test_agent_metrics_session, expected_series, seq_id=2)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="mock.ANY doesn't works in py3.5 or lower")
def test_send_metric_datapoint_equal_type_and_tags_yields_single_series(test_agent_metrics_session, mock_time):
    """Check metrics datapoints and the aggregations by datapoint ID.
    A datapoint ID is at least: a metric name, a metric value, and the time at which the value was collected.
    But in Datadog, a datapoint also includes tags, which declare all the various scopes the datapoint belongs to
    https://www.datadoghq.com/blog/the-power-of-tagged-metrics/#whats-a-metric-tag
    """
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 2, {"a": "b"})
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 3, {"a": "b"})

        expected_series = [
            {
                "common": True,
                "metric": "test-metric",
                "points": [[1642544540, 5.0]],
                "tags": ["a:b"],
                "type": "count",
            },
        ]

        _assert_metric(test_agent_metrics_session, expected_series)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="mock.ANY doesn't works in py3.5 or lower")
def test_send_metric_datapoint_equal_type_different_tags_yields_multiple_series(test_agent_metrics_session, mock_time):
    """Check metrics datapoints and the aggregations by datapoint ID.
    A datapoint ID is at least: a metric name, a metric value, and the time at which the value was collected.
    But in Datadog, a datapoint also includes tags, which declare all the various scopes the datapoint belongs to
    https://www.datadoghq.com/blog/the-power-of-tagged-metrics/#whats-a-metric-tag
    """
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 4, {"a": "b"})
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 5, {"a": "b", "c": "d"})
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 6, {})

        expected_series = [
            {
                "common": True,
                "metric": "test-metric",
                "points": [[1642544540, 4.0]],
                "tags": ["a:b"],
                "type": "count",
            },
            {
                "common": True,
                "metric": "test-metric",
                "points": [[1642544540, 5.0]],
                "tags": ["a:b", "c:d"],
                "type": "count",
            },
            {
                "common": True,
                "metric": "test-metric",
                "points": [[1642544540, 6.0]],
                "tags": [],
                "type": "count",
            },
        ]

        _assert_metric(test_agent_metrics_session, expected_series)


def test_send_metric_datapoint_equal_tags_different_type_throws_error(test_agent_metrics_session, mock_time):
    """Check metrics datapoints and the aggregations by datapoint ID.
    A datapoint ID is at least: a metric name, a metric value, and the time at which the value was collected.
    But in Datadog, a datapoint also includes tags, which declare all the various scopes the datapoint belongs to
    https://www.datadoghq.com/blog/the-power-of-tagged-metrics/#whats-a-metric-tag
    """
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 1, {"a": "b"})
        with pytest.raises(TelemetryTypeError) as e:
            telemetry_writer.add_gauge_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 1, {"a": "b"})

            assert e.value.args[0] == (
                'Error: metric with name "test-metric" and type "count" '
                'exists. You can\'t create a new metric with this name an type "gauge"'
            )


@pytest.mark.skipif(sys.version_info < (3, 6), reason="mock.ANY doesn't works in py3.5 or lower")
def test_send_tracers_count_metric(test_agent_metrics_session, mock_time):
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 1, {"a": "b"})
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 1, {"a": "b"})
        telemetry_writer.add_count_metric(TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 1, {})
        telemetry_writer.add_count_metric(
            TELEMETRY_NAMESPACE_TAG_TRACER, "test-metric", 1, {"hi": "HELLO", "NAME": "CANDY"}
        )

        expected_series = [
            {
                "common": True,
                "metric": "test-metric",
                "points": [[1642544540, 2.0]],
                "tags": ["a:b"],
                "type": "count",
            },
            {
                "common": True,
                "metric": "test-metric",
                "points": [[1642544540, 1.0]],
                "tags": [],
                "type": "count",
            },
            {
                "common": True,
                "metric": "test-metric",
                "points": [[1642544540, 1.0]],
                "tags": ["hi:HELLO", "NAME:CANDY"],
                "type": "count",
            },
        ]
        _assert_metric(test_agent_metrics_session, expected_series)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="mock.ANY doesn't works in py3.5 or lower")
def test_send_appsec_rate_metric(test_agent_metrics_session, mock_time):
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_rate_metric(
            TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 1, {"hi": "HELLO", "NAME": "CANDY"}
        )
        telemetry_writer.add_rate_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 1, {})
        telemetry_writer.add_rate_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 1, {})

        expected_series = [
            {
                "common": True,
                "interval": 60,
                "metric": "test-metric",
                "points": [[1642544540, 0.016666666666666666]],
                "tags": ["hi:HELLO", "NAME:CANDY"],
                "type": "rate",
            },
            {
                "common": True,
                "interval": 60,
                "metric": "test-metric",
                "points": [[1642544540, 0.03333333333333333]],
                "tags": [],
                "type": "rate",
            },
        ]

        _assert_metric(test_agent_metrics_session, expected_series, namespace=TELEMETRY_NAMESPACE_TAG_APPSEC)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="mock.ANY doesn't works in py3.5 or lower")
def test_send_appsec_gauge_metric(test_agent_metrics_session, mock_time):
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_gauge_metric(
            TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 5, {"hi": "HELLO", "NAME": "CANDY"}
        )
        telemetry_writer.add_gauge_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 5, {"a": "b"})
        telemetry_writer.add_gauge_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 6, {})

        expected_series = [
            {
                "common": True,
                "interval": 60,
                "metric": "test-metric",
                "points": [[1642544540, 5.0]],
                "tags": ["hi:HELLO", "NAME:CANDY"],
                "type": "gauge",
            },
            {
                "common": True,
                "interval": 60,
                "metric": "test-metric",
                "points": [[1642544540, 5.0]],
                "tags": ["a:b"],
                "type": "gauge",
            },
            {
                "common": True,
                "interval": 60,
                "metric": "test-metric",
                "points": [[1642544540, 6.0]],
                "tags": [],
                "type": "gauge",
            },
        ]
        _assert_metric(test_agent_metrics_session, expected_series, namespace=TELEMETRY_NAMESPACE_TAG_APPSEC)


@pytest.mark.skipif(sys.version_info < (3, 6), reason="mock.ANY doesn't works in py3.5 or lower")
def test_send_appsec_distributions_metric(test_agent_metrics_session, mock_time):
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_distribution_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 4, {})
        telemetry_writer.add_distribution_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 5, {})
        telemetry_writer.add_distribution_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 6, {})

        expected_series = [
            {
                "metric": "test-metric",
                "points": [4.0, 5.0, 6.0],
                "tags": [],
            }
        ]
        _assert_metric(
            test_agent_metrics_session,
            expected_series,
            namespace=TELEMETRY_NAMESPACE_TAG_APPSEC,
            type_paypload=TELEMETRY_TYPE_DISTRIBUTION,
        )


@pytest.mark.skipif(sys.version_info < (3, 6), reason="mock.ANY doesn't works in py3.5 or lower")
def test_send_metric_flush_and_distributions_series_is_restarted(test_agent_metrics_session, mock_time):
    """Check the queue of metrics is empty after run periodic method of PeriodicService"""
    with override_global_config(dict(_telemetry_metrics_enabled=True)):
        telemetry_writer = test_agent_metrics_session.telemetry_writer
        telemetry_writer.add_distribution_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 4, {})
        telemetry_writer.add_distribution_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 5, {})
        telemetry_writer.add_distribution_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 6, {})
        expected_series = [
            {
                "metric": "test-metric",
                "points": [4.0, 5.0, 6.0],
                "tags": [],
            }
        ]

        _assert_metric(
            test_agent_metrics_session,
            expected_series,
            namespace=TELEMETRY_NAMESPACE_TAG_APPSEC,
            type_paypload=TELEMETRY_TYPE_DISTRIBUTION,
        )

        expected_series = [
            {
                "metric": "test-metric",
                "points": [1.0],
                "tags": [],
            }
        ]

        telemetry_writer.add_distribution_metric(TELEMETRY_NAMESPACE_TAG_APPSEC, "test-metric", 1, {})

        _assert_metric(
            test_agent_metrics_session,
            expected_series,
            namespace=TELEMETRY_NAMESPACE_TAG_APPSEC,
            type_paypload=TELEMETRY_TYPE_DISTRIBUTION,
            seq_id=2,
        )



import itertools
import os
from typing import ClassVar
from typing import Optional
from typing import Set

import attr

import ddtrace
from ddtrace.internal import atexit
from ddtrace.internal import forksafe

from .. import periodic
from ..dogstatsd import get_dogstatsd_client
from ..logger import get_logger
from .constants import DEFAULT_RUNTIME_METRICS
from .constants import DEFAULT_RUNTIME_TAGS
from .metric_collectors import GCRuntimeMetricCollector
from .metric_collectors import PSUtilRuntimeMetricCollector
from .tag_collectors import PlatformTagCollector
from .tag_collectors import TracerTagCollector


log = get_logger(__name__)


class RuntimeCollectorsIterable(object):
    def __init__(self, enabled=None):
        self._enabled = enabled or self.ENABLED
        # Initialize the collectors.
        self._collectors = [c() for c in self.COLLECTORS]

    def __iter__(self):
        collected = (collector.collect(self._enabled) for collector in self._collectors)
        return itertools.chain.from_iterable(collected)

    def __repr__(self):
        return "{}(enabled={})".format(
            self.__class__.__name__,
            self._enabled,
        )


class RuntimeTags(RuntimeCollectorsIterable):
    ENABLED = DEFAULT_RUNTIME_TAGS
    COLLECTORS = [
        PlatformTagCollector,
        TracerTagCollector,
    ]


class RuntimeMetrics(RuntimeCollectorsIterable):
    ENABLED = DEFAULT_RUNTIME_METRICS
    COLLECTORS = [
        GCRuntimeMetricCollector,
        PSUtilRuntimeMetricCollector,
    ]


def _get_interval_or_default():
    return float(os.getenv("DD_RUNTIME_METRICS_INTERVAL", default=10))


@attr.s(eq=False)
class RuntimeWorker(periodic.PeriodicService):
    """Worker thread for collecting and writing runtime metrics to a DogStatsd
    client.
    """

    _interval = attr.ib(type=float, factory=_get_interval_or_default)
    tracer = attr.ib(type=ddtrace.Tracer, default=None)
    dogstatsd_url = attr.ib(type=Optional[str], default=None)
    _dogstatsd_client = attr.ib(init=False, repr=False)
    _runtime_metrics = attr.ib(factory=RuntimeMetrics, repr=False)
    _services = attr.ib(type=Set[str], init=False, factory=set)

    enabled = False
    _instance = None  # type: ClassVar[Optional[RuntimeWorker]]
    _lock = forksafe.Lock()

    def __attrs_post_init__(self):
        # type: () -> None
        self._dogstatsd_client = get_dogstatsd_client(self.dogstatsd_url or ddtrace.internal.agent.get_stats_url())
        self.tracer = self.tracer or ddtrace.tracer

    @classmethod
    def disable(cls):
        # type: () -> None
        with cls._lock:
            if cls._instance is None:
                return

            forksafe.unregister(cls._restart)

            cls._instance.stop()
            # DEV: Use timeout to avoid locking on shutdown. This seems to be
            # required on some occasions by Python 2.7. Deadlocks seem to happen
            # when some functionalities (e.g. platform.architecture) are used
            # which end up calling
            #    _execute_child (/usr/lib/python2.7/subprocess.py:1023)
            # This is a continuous attempt to read:
            #    _eintr_retry_call (/usr/lib/python2.7/subprocess.py:125)
            # which is the eventual cause of the deadlock.
            cls._instance.join(1)
            cls._instance = None
            cls.enabled = False

    @classmethod
    def _restart(cls):
        cls.disable()
        cls.enable()

    @classmethod
    def enable(cls, flush_interval=None, tracer=None, dogstatsd_url=None):
        # type: (Optional[float], Optional[ddtrace.Tracer], Optional[str]) -> None
        with cls._lock:
            if cls._instance is not None:
                return
            if flush_interval is None:
                flush_interval = _get_interval_or_default()
            runtime_worker = cls(flush_interval, tracer, dogstatsd_url)  # type: ignore[arg-type]
            runtime_worker.start()
            # force an immediate update constant tags
            runtime_worker.update_runtime_tags()

            forksafe.register(cls._restart)
            atexit.register(cls.disable)

            cls._instance = runtime_worker
            cls.enabled = True

    def flush(self):
        # type: () -> None
        # The constant tags for the dogstatsd client needs to updated with any new
        # service(s) that may have been added.
        if self._services != self.tracer._services:
            self._services = self.tracer._services
            self.update_runtime_tags()

        with self._dogstatsd_client:
            for key, value in self._runtime_metrics:
                log.debug("Writing metric %s:%s", key, value)
                self._dogstatsd_client.distribution(key, value)

    def _stop_service(self):
        # type: (...) -> None
        # De-register span hook
        super(RuntimeWorker, self)._stop_service()

    def update_runtime_tags(self):
        # type: () -> None
        # DEV: ddstatsd expects tags in the form ['key1:value1', 'key2:value2', ...]
        tags = ["{}:{}".format(k, v) for k, v in RuntimeTags()]
        log.debug("Updating constant tags %s", tags)
        self._dogstatsd_client.constant_tags = tags

    periodic = flush
    on_shutdown = flush



from typing import Any
from typing import Callable
from typing import List
from typing import Optional
from typing import Tuple

from ddtrace.debugging._capture.model import CaptureState
from ddtrace.debugging._capture.model import CapturedEvent
from ddtrace.debugging._encoding import BufferedEncoder
from ddtrace.debugging._metrics import metrics
from ddtrace.internal._encoding import BufferFull
from ddtrace.internal.compat import ExcInfoType
from ddtrace.internal.logger import get_logger


CaptorType = Callable[[List[Tuple[str, Any]], List[Tuple[str, Any]], ExcInfoType, int], Any]

log = get_logger(__name__)
meter = metrics.get_meter("snapshot.collector")


NO_RETURN_VALUE = object()


class CapturedEventWithContext(object):
    """CapturedEvent with context.

    This is used to capture event for function invocation. the CapturedEventWithContext
    would call the event.enter() and event.exit() when the function starts and ends.

    The handler is triggered just after event.exit() is completed.
    """

    def __init__(
        self,
        event,  # type: CapturedEvent
        handler,  # type: Callable[[CapturedEvent],None]
    ):
        # type: (...) -> None
        self._on_exit_handler = handler
        self.event = event
        self.return_value = NO_RETURN_VALUE  # type: Any
        self.duration = None  # type: Optional[int]

        self.event.enter()

    def exit(self, retval, exc_info, duration_ns):
        # type: (Any, ExcInfoType, int) -> None
        """Exit the snapshot context.

        The arguments are used to record either the return value or the exception, and
        the duration of the wrapped call.
        """
        self.return_value = retval
        self.duration = duration_ns

        return self.__exit__(*exc_info)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        # type: (ExcInfoType) -> None
        self.event.exit(self.return_value, exc_info, self.duration)
        self._on_exit_handler(self.event)


class CapturedEventCollector(object):
    """Captured Event collector.

    This is used to collect and encode snapshot information as soon as
    requested. The ``push`` method is intended to be called in point
    event is over and information is already available and ready to be encoded
    or event status indicate it should be skipped.
    For function instrumentation (e.g. function probes), we use the ``attach`` method to create a
    ``CapturedEventWithContext`` instance that can be used to capture additional data,
    such as the return value of the wrapped function.
    """

    def __init__(self, encoder):
        # type: (BufferedEncoder) -> None
        self._encoder = encoder

    def _enqueue(self, snapshot):
        # type: (Any) -> None
        try:
            self._encoder.put(snapshot)
        except BufferFull:
            log.debug("Encoder buffer full")
            meter.increment("encoder.buffer.full")

    def push(self, event):
        # type: (CapturedEvent) -> None
        if event.state == CaptureState.SKIP_COND:
            meter.increment("skip", tags={"cause": "cond", "probe_id": event.probe.probe_id})
        elif event.state == CaptureState.SKIP_COND_ERROR:
            meter.increment("skip", tags={"cause": "cond_error", "probe_id": event.probe.probe_id})
        elif event.state == CaptureState.COND_ERROR_AND_COMMIT:
            meter.increment("skip", tags={"cause": "cond_error", "probe_id": event.probe.probe_id})
            self._enqueue(event)
        elif event.state == CaptureState.SKIP_RATE:
            meter.increment("skip", tags={"cause": "rate", "probe_id": event.probe.probe_id})
        elif event.state == CaptureState.DONE_AND_COMMIT:
            meter.increment("capture", tags={"probe_id": event.probe.probe_id})
            self._enqueue(event)

    def attach(self, event):
        # type: (CapturedEvent) -> CapturedEventWithContext
        """Collect via a snapshot context."""
        return CapturedEventWithContext(event, lambda e: self.push(e))



import pytest

from ddtrace.appsec import _asm_request_context
from ddtrace.contrib.trace_utils import set_http_meta
from ddtrace.ext import SpanTypes
from ddtrace.internal.telemetry import telemetry_metrics_writer
from ddtrace.internal.telemetry.constants import TELEMETRY_NAMESPACE_TAG_APPSEC
from ddtrace.internal.telemetry.constants import TELEMETRY_TYPE_DISTRIBUTION
from ddtrace.internal.telemetry.constants import TELEMETRY_TYPE_GENERATE_METRICS
from tests.appsec.test_processor import Config
from tests.appsec.test_processor import RULES_GOOD_PATH
from tests.appsec.test_processor import _BLOCKED_IP
from tests.appsec.test_processor import _enable_appsec
from tests.utils import override_env
from tests.utils import override_global_config


@pytest.fixture
def mock_telemetry_metrics_writer():
    metrics_result = telemetry_metrics_writer._namespace._metrics_data
    assert len(metrics_result[TELEMETRY_TYPE_GENERATE_METRICS][TELEMETRY_NAMESPACE_TAG_APPSEC]) == 0
    assert len(metrics_result[TELEMETRY_TYPE_DISTRIBUTION][TELEMETRY_NAMESPACE_TAG_APPSEC]) == 0
    yield telemetry_metrics_writer
    telemetry_metrics_writer._flush_namespace_metrics()


def _assert_generate_metrics(metrics_result, is_rule_triggered=False, is_blocked_request=False):
    generate_metrics = metrics_result[TELEMETRY_TYPE_GENERATE_METRICS][TELEMETRY_NAMESPACE_TAG_APPSEC]
    assert len(generate_metrics) == 2, "Expected 2 generate_metrics"
    for metric_id, metric in generate_metrics.items():
        if metric.name == "waf.requests":
            assert metric._tags["rule_triggered"] is is_rule_triggered
            assert metric._tags["request_blocked"] is is_blocked_request
            assert len(metric._tags["waf_version"]) > 0
            assert len(metric._tags["event_rules_version"]) > 0
        elif metric.name == "event_rules.loaded":
            pass
        else:
            pytest.fail("Unexpected generate_metrics {}".format(metric.name))


def _assert_distributions_metrics(metrics_result, is_rule_triggered=False, is_blocked_request=False):
    distributions_metrics = metrics_result[TELEMETRY_TYPE_DISTRIBUTION][TELEMETRY_NAMESPACE_TAG_APPSEC]

    assert len(distributions_metrics) == 2, "Expected 2 distributions_metrics"
    for metric_id, metric in distributions_metrics.items():
        if metric.name in ["waf.duration", "waf.duration_ext"]:
            assert len(metric._points) >= 1
            assert type(metric._points[0]) is float
            assert metric._tags["rule_triggered"] is is_rule_triggered
            assert metric._tags["request_blocked"] is is_blocked_request
            assert len(metric._tags["waf_version"]) > 0
            assert len(metric._tags["event_rules_version"]) > 0
        else:
            pytest.fail("Unexpected distributions_metrics {}".format(metric.name))


def test_metrics_when_appsec_doesnt_runs(mock_telemetry_metrics_writer, tracer):
    with override_global_config(dict(_appsec_enabled=False, _telemetry_metrics_enabled=True)):
        tracer.configure(api_version="v0.4", appsec_enabled=False)
        mock_telemetry_metrics_writer._flush_namespace_metrics()
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(
                span,
                Config(),
            )
    metrics_data = mock_telemetry_metrics_writer._namespace._metrics_data
    assert len(metrics_data[TELEMETRY_TYPE_GENERATE_METRICS][TELEMETRY_NAMESPACE_TAG_APPSEC]) == 0
    assert len(metrics_data[TELEMETRY_TYPE_DISTRIBUTION][TELEMETRY_NAMESPACE_TAG_APPSEC]) == 0


def test_metrics_when_appsec_runs(mock_telemetry_metrics_writer, tracer):
    with override_global_config(dict(_appsec_enabled=True, _telemetry_metrics_enabled=True)):
        _enable_appsec(tracer)
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(
                span,
                Config(),
            )
    _assert_generate_metrics(mock_telemetry_metrics_writer._namespace._metrics_data)
    _assert_distributions_metrics(mock_telemetry_metrics_writer._namespace._metrics_data)


def test_metrics_when_appsec_attack(mock_telemetry_metrics_writer, tracer):
    with override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)), override_global_config(
        dict(_appsec_enabled=True, _telemetry_metrics_enabled=True)
    ):
        _enable_appsec(tracer)
        with tracer.trace("test", span_type=SpanTypes.WEB) as span:
            set_http_meta(span, Config(), request_cookies={"attack": "1' or '1' = '1'"})
    _assert_generate_metrics(mock_telemetry_metrics_writer._namespace._metrics_data, is_rule_triggered=True)
    _assert_distributions_metrics(mock_telemetry_metrics_writer._namespace._metrics_data, is_rule_triggered=True)


def test_metrics_when_appsec_block(mock_telemetry_metrics_writer, tracer):
    with override_env(dict(DD_APPSEC_RULES=RULES_GOOD_PATH)), override_global_config(
        dict(_appsec_enabled=True, _telemetry_metrics_enabled=True)
    ):
        _enable_appsec(tracer)
        with _asm_request_context.asm_request_context_manager(_BLOCKED_IP, {}):
            with tracer.trace("test", span_type=SpanTypes.WEB) as span:
                set_http_meta(
                    span,
                    Config(),
                )

    _assert_generate_metrics(
        mock_telemetry_metrics_writer._namespace._metrics_data, is_rule_triggered=True, is_blocked_request=True
    )
    _assert_distributions_metrics(
        mock_telemetry_metrics_writer._namespace._metrics_data, is_rule_triggered=True, is_blocked_request=True
    )



# This test script was automatically generated by the contrib-patch-tests.py
# script. If you want to make changes to it, you should make sure that you have
# removed the ``_generated`` suffix from the file name, to prevent the content
# from being overwritten by future re-generations.

from ddtrace.contrib.boto.patch import patch


try:
    from ddtrace.contrib.boto.patch import unpatch
except ImportError:
    unpatch = None
from tests.contrib.patch import PatchTestCase


class TestBotoPatch(PatchTestCase.Base):
    __integration_name__ = "boto"
    __module_name__ = "boto.connection"
    __patch_func__ = patch
    __unpatch_func__ = unpatch

    def assert_module_patched(self, boto_connection):
        pass

    def assert_not_module_patched(self, boto_connection):
        pass

    def assert_not_module_double_patched(self, boto_connection):
        pass



#!/usr/bin/env python3

import ast
import codecs
import os
from types import ModuleType
from typing import Optional
from typing import Tuple

from ddtrace.appsec.iast._ast.visitor import AstVisitor
from ddtrace.internal.logger import get_logger
from ddtrace.internal.module import origin


# Prefixes for modules where IAST patching is allowed
IAST_ALLOWLIST = ("tests.appsec.iast",)

ENCODING = ""

log = get_logger(__name__)


def get_encoding(module_path):  # type: (str) -> str
    """
    First tries to detect the encoding for the file,
    otherwise, returns global encoding default
    """
    global ENCODING
    if not ENCODING:
        try:
            ENCODING = codecs.lookup("utf-8-sig").name
        except LookupError:
            ENCODING = codecs.lookup("utf-8").name
    return ENCODING


def _should_iast_patch(module_name):
    return not module_name.startswith("ddtrace") and module_name.startswith(IAST_ALLOWLIST)


def visit_ast(
    source_text,  # type: str
    module_path,  # type: str
    module_name="",  # type: str
):  # type: (...) -> Optional[str]
    parsed_ast = ast.parse(source_text, module_path)

    visitor = AstVisitor(
        filename=module_path,
        module_name=module_name,
    )
    modified_ast = visitor.visit(parsed_ast)

    if not visitor.ast_modified:
        return None

    return modified_ast


def astpatch_module(module):  # type: (ModuleType) -> Tuple[str, str]
    module_name = module.__name__
    module_path = origin(module)
    try:
        if os.stat(module_path).st_size == 0:
            # Don't patch empty files like __init__.py
            log.debug("empty file: %s", module_path)
            return "", ""
    except OSError:
        log.debug("astpatch_source couldn't find the file: %s", module_path, exc_info=True)
        return "", ""

    # Get the file extension, if it's dll, os, pyd, dyn, dynlib: return
    # If its pyc or pyo, change to .py and check that the file exists. If not,
    # return with warning.
    _, module_ext = os.path.splitext(module_path)

    if module_ext.lower() not in {".pyo", ".pyc", ".pyw", ".py"}:
        # Probably native or built-in module
        log.debug("extension not supported: %s for: %s", module_ext, module_path)
        return "", ""

    with open(module_path, "r", encoding=get_encoding(module_path)) as source_file:
        try:
            source_text = source_file.read()
        except UnicodeDecodeError:
            log.debug("unicode decode error for file: %s", module_path, exc_info=True)
            return "", ""

    if len(source_text.strip()) == 0:
        # Don't patch empty files like __init__.py
        log.debug("empty file: %s", module_path)
        return "", ""

    new_source = visit_ast(
        source_text,
        module_path,
        module_name=module_name,
    )
    if new_source is None:
        log.debug("file not ast patched: %s", module_path)
        return "", ""

    return module_path, new_source



# tags
NAME = "db.name"  # the database name (eg: dbname for pgsql)
USER = "db.user"  # the user connecting to the db
SYSTEM = "db.system"  # the database's DBMS name (e.g. postgresql for pgsql)
ROWCOUNT = "db.row_count"  # the rowcount of a query
SYSTEM = "db.system"  # the database's DBMS name (e.g. postgresql for pgsql)



#!/usr/bin/env python3

from builtins import str as builtin_str

from ddtrace.appsec.iast._input_info import Input_info
from ddtrace.appsec.iast._taint_tracking import add_taint_pyobject  # type: ignore[attr-defined]
from ddtrace.appsec.iast._taint_tracking import is_pyobject_tainted  # type: ignore[attr-defined]
from ddtrace.appsec.iast._taint_tracking import taint_pyobject  # type: ignore[attr-defined]


def str_aspect(*args, **kwargs):
    result = builtin_str(*args, **kwargs)
    if isinstance(args[0], (str, bytes, bytearray)) and is_pyobject_tainted(args[0]):
        result = taint_pyobject(result, Input_info("str_aspect", result, 0))

    return result


def add_aspect(op1, op2):
    if not isinstance(op1, (str, bytes, bytearray)):
        return op1 + op2

    return add_taint_pyobject(getattr(op1.__class__, "__add__")(op1, op2), op1, op2)



"""
Generate missing integration patch tests.

This scripts scans the ddtrace/contrib folder for all the currently supported
integrations that implement the patch method in the patch submodule, and checks
that there is a corresponding patch test is implemented in tests/contrib. It
then generates a basic patch test for those integrations that do not have one.

Run as

    python scripts/contrib-patch-tests.py

from the repository root folder to automatically (re-)generate missing
integration patch tests.
"""

import ast
from pathlib import Path
from textwrap import dedent


CONTRIB_PATH = Path.cwd() / "ddtrace" / "contrib"
CONTRIB_TEST_PATH = Path().cwd() / "tests" / "contrib"


def generate_patch_test_source(contrib):
    print(f"Generating patch test for {contrib}")

    # Parse the contrib module to get the ``required_modules`` list
    # DEV: Parsing the module is much lighter and safer than importing it
    module_ast = ast.parse((CONTRIB_PATH / contrib / "__init__.py").read_text())
    required_modules_node = [
        _ for _ in module_ast.body if isinstance(_, ast.Assign) and _.targets[0].id == "required_modules"
    ]
    if not required_modules_node:
        return

    required_modules = [_.value for _ in required_modules_node[0].value.elts]

    if not required_modules:
        return

    # We use the first entry in the ``required_modules`` for the module name.
    module = required_modules[0]

    (CONTRIB_TEST_PATH / contrib / f"test_{contrib}_patch_generated.py").write_text(
        dedent(
            f"""
            # This test script was automatically generated by the contrib-patch-tests.py
            # script. If you want to make changes to it, you should make sure that you have
            # removed the ``_generated`` suffix from the file name, to prevent the content
            # from being overwritten by future re-generations.

            from ddtrace.contrib.{contrib}.patch import patch


            try:
                from ddtrace.contrib.{contrib}.patch import unpatch
            except ImportError:
                unpatch = None
            from tests.contrib.patch import PatchTestCase


            class Test{contrib.title()}Patch(PatchTestCase.Base):
                __integration_name__ = "{contrib}"
                __module_name__ = "{module}"
                __patch_func__ = patch
                __unpatch_func__ = unpatch

                def assert_module_patched(self, {module.replace(".", "_")}):
                    pass

                def assert_not_module_patched(self, {module.replace(".", "_")}):
                    pass

                def assert_not_module_double_patched(self, {module.replace(".", "_")}):
                    pass
            """
        ).lstrip()
    )


# Take all the contribs that have a patch submodule. We assume they implement
# the patch API by exposing a ``patch`` function.
for contrib in [p.parent.name for p in CONTRIB_PATH.glob("*/patch.py")]:
    if not any((CONTRIB_TEST_PATH / contrib / f).is_file() for f in ("test_patch.py", f"test_{contrib}_patch.py")):
        generate_patch_test_source(contrib)



import os
import signal
import subprocess
import time
from typing import Callable
from typing import Dict
from typing import Generator
from typing import List

import pytest
import six
import tenacity

from ddtrace.contrib.flask.patch import flask_version
from ddtrace.internal.constants import APPSEC_BLOCKED_RESPONSE_HTML
from ddtrace.internal.constants import APPSEC_BLOCKED_RESPONSE_JSON
from tests.appsec.test_processor import RULES_GOOD_PATH
from tests.appsec.test_processor import _BLOCKED_IP
from tests.contrib.flask.test_flask_appsec import _ALLOWED_USER
from tests.contrib.flask.test_flask_appsec import _BLOCKED_USER
from tests.webclient import Client


DEFAULT_HEADERS = {
    "User-Agent": "python-httpx/x.xx.x",
}


@pytest.fixture
def flask_port():
    # type: () -> str
    return "8000"


@pytest.fixture
def flask_wsgi_application():
    # type: () -> str
    return "tests.contrib.flask.app:app"


@pytest.fixture
def flask_command(flask_wsgi_application, flask_port):
    # type: (str, str) -> List[str]
    cmd = "ddtrace-run flask run -h 0.0.0.0 -p %s" % (flask_port,)
    return cmd.split()


def flask_appsec_good_rules_env(flask_wsgi_application):
    # type: (str) -> Dict[str, str]
    env = os.environ.copy()
    env.update(
        {
            # Avoid noisy database spans being output on app startup/teardown.
            "DD_TRACE_SQLITE3_ENABLED": "0",
            "FLASK_APP": flask_wsgi_application,
            "DD_APPSEC_ENABLED": "true",
            "DD_APPSEC_RULES": RULES_GOOD_PATH,
        }
    )
    return env


@pytest.fixture
def flask_client(flask_command, flask_port, flask_wsgi_application, flask_env_arg):
    # type: (List[str], Dict[str, str], str, Callable) -> Generator[Client, None, None]
    # Copy the env to get the correct PYTHONPATH and such
    # from the virtualenv.
    # webservers might exec or fork into another process, so we need to os.setsid() to create a process group
    # (all of which will listen to signals sent to the parent) so that we can kill the whole application.
    proc = subprocess.Popen(
        flask_command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
        env=flask_env_arg(flask_wsgi_application),
        preexec_fn=os.setsid,
        cwd=str(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))),
    )
    try:
        client = Client("http://0.0.0.0:%s" % flask_port)
        # Wait for the server to start up
        try:
            client.wait()
        except tenacity.RetryError:
            # process failed
            stdout = proc.stdout.read()
            stderr = proc.stderr.read()
            raise TimeoutError(
                "Server failed to start\n======STDOUT=====%s\n\n======STDERR=====%s\n" % (stdout, stderr)
            )
        yield client
        try:
            client.get_ignored("/shutdown")
        except Exception:
            pass
        # At this point the traces have been sent to the test agent
        # but the test agent hasn't necessarily finished processing
        # the traces (race condition) so wait just a bit for that
        # processing to complete.
        time.sleep(0.2)
    finally:
        os.killpg(proc.pid, signal.SIGKILL)
        proc.wait()
    # DEV uncomment those lines if you need more info locally
    # stdout = proc.stdout.read()
    # print(stdout)
    # stderr = proc.stderr.read()
    # print(stderr.decode("UTF-8"))


@pytest.mark.snapshot(
    ignores=[
        "meta._dd.appsec.waf.duration",
        "meta._dd.appsec.waf.duration_ext",
        "meta.flask.version",
        "meta.http.request.headers.accept-encoding",
        "meta.http.request.headers.user-agent",
        "http.response.headers.content-length",
        "http.response.headers.content-type",
        "meta.http.useragent",
        "meta.error.stack",
        "metrics._dd.appsec.event_rules.loaded",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
        "meta.span.kind",
    ],
    variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)},
)
@pytest.mark.parametrize("flask_env_arg", (flask_appsec_good_rules_env,))
def test_flask_ipblock_match_403(flask_client):
    resp = flask_client.get("/", headers={"X-Real-Ip": _BLOCKED_IP, "ACCEPT": "text/html"})
    assert resp.status_code == 403
    if hasattr(resp, "text"):
        assert resp.text == APPSEC_BLOCKED_RESPONSE_HTML
    else:
        assert resp.data == six.ensure_binary(APPSEC_BLOCKED_RESPONSE_HTML)


@pytest.mark.snapshot(
    ignores=[
        "meta._dd.appsec.waf.duration",
        "meta._dd.appsec.waf.duration_ext",
        "meta.flask.version",
        "meta.http.request.headers.accept-encoding",
        "meta.http.request.headers.user-agent",
        "http.response.headers.content-length",
        "http.response.headers.content-type",
        "meta.http.useragent",
        "meta.error.stack",
        "metrics._dd.appsec.event_rules.loaded",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
        "meta.span.kind",
    ],
    variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)},
)
@pytest.mark.parametrize("flask_env_arg", (flask_appsec_good_rules_env,))
def test_flask_ipblock_match_403_json(flask_client):
    resp = flask_client.get("/", headers={"X-Real-Ip": _BLOCKED_IP})
    assert resp.status_code == 403
    if hasattr(resp, "text"):
        assert resp.text == APPSEC_BLOCKED_RESPONSE_JSON
    else:
        assert resp.data == six.ensure_binary(APPSEC_BLOCKED_RESPONSE_JSON)


@pytest.mark.snapshot(
    ignores=[
        "meta._dd.appsec.waf.duration",
        "meta._dd.appsec.waf.duration_ext",
        "meta.flask.version",
        "meta.http.request.headers.accept-encoding",
        "meta.http.request.headers.user-agent",
        "http.response.headers.content-length",
        "http.response.headers.content-type",
        "meta.http.useragent",
        "meta.error.stack",
        "metrics._dd.appsec.event_rules.loaded",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ],
    variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)},
)
@pytest.mark.parametrize("flask_env_arg", (flask_appsec_good_rules_env,))
def test_flask_userblock_match_403_json(flask_client):
    resp = flask_client.get("/checkuser/%s" % _BLOCKED_USER)
    assert resp.status_code == 403
    if hasattr(resp, "text"):
        assert resp.text == APPSEC_BLOCKED_RESPONSE_JSON
    else:
        assert resp.data == six.ensure_binary(APPSEC_BLOCKED_RESPONSE_JSON)


@pytest.mark.snapshot(
    ignores=[
        "meta._dd.appsec.waf.duration",
        "meta._dd.appsec.waf.duration_ext",
        "meta.flask.version",
        "meta.http.request.headers.accept-encoding",
        "meta.http.request.headers.user-agent",
        "http.response.headers.content-length",
        "http.response.headers.content-type",
        "meta.http.useragent",
        "meta.error.stack",
        "metrics._dd.appsec.event_rules.loaded",
        "metrics._dd.appsec.waf.duration",
        "metrics._dd.appsec.waf.duration_ext",
    ],
    variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)},
)
@pytest.mark.parametrize("flask_env_arg", (flask_appsec_good_rules_env,))
def test_flask_userblock_match_200_json(flask_client):
    resp = flask_client.get("/checkuser/%s" % _ALLOWED_USER)
    assert resp.status_code == 200
